################################################################################

import sys
import types
from types import GenericAlias, TracebackType, CodeType
from typing import (
    TypeVarTuple,
    ClassVar,
    AbstractSet,
    MutableMapping,
    Iterable,
    Iterator,
    Callable,
    overload,
    SupportsBytes,
    Any,
    TypeVar,
    SupportsAbs,
    SupportsComplex,
    Sized,
    IO,
    Literal,
    BinaryIO,
    Reversible,
)
from collections.abc import Iterable, Buffer, Iterator, Reversible, Callable, Sized, MutableMapping
from io import FileIO, BufferedReader, BufferedWriter, BufferedRandom, TextIOWrapper
from os import PathLike
import ast
from ._typeshed_init import (
    SupportsIter,
    SupportsDunderLT,
    SupportsNext,
    SupportsWrite,
    SupportsKeysAndGetItem,
    SupportsDunderGT,
    SupportsLenAndGetItem,
    SupportsDivMod,
    SupportsRDivMod,
)
from ._dict_extra import dict_keys, dict_items, dict_values

################################################################################

__classes__ = [
    '_object',
    '_staticmethod',
    '_classmethod',
    '_type',
    '_super',
    '_int',
    '_float',
    '_complex',
    '_FormatMapMapping',
    '_TranslateTable',
    '_str',
    '_bytes',
    '_bytearray',
    '_memoryview',
    '_bool',
    '_slice',
    '_tuple',
    'function',
    '_list',
    '_dict',
    '_set',
    '_frozenset',
    '_enumerate',
    '_range',
    '_property',
    '_NotImplementedType',
    '_filter',
    '_GetItemIterable',
    '_map',
    '_SupportsWriteAndFlush',
    '_SupportsPow2',
    '_SupportsPow3NoneOnly',
    '_SupportsPow3',
    '_reversed',
    '_SupportsRound1',
    '_SupportsRound2',
    '_SupportsSumWithNoDefaultGiven',
    '_BaseException',
    '_GeneratorExit',
    '_KeyboardInterrupt',
    '_SystemExit',
    '_Exception',
    '_StopIteration',
    '_StopAsyncIteration',
]

__functions__ = [
    '_abs',
    '_all',
    '_any',
    '_ascii',
    '_bin',
    '_breakpoint',
    '_callable',
    '_chr',
    '_compile',
    '_compile',
    '_compile',
    '_compile',
    '_delattr',
    '_dir',
    '_divmod',
    '_divmod',
    '_format',
    '_getattr',
    '_getattr',
    '_getattr',
    '_getattr',
    '_getattr',
    '_getattr',
    '_globals',
    '_hasattr',
    '_hash',
    '_hex',
    '_id',
    '_input',
    '_iter',
    '_iter',
    '_iter',
    '_iter',
    '_isinstance',
    '_issubclass',
    '_len',
    '_locals',
    '_max',
    '_max',
    '_max',
    '_max',
    '_max',
    '_max',
    '_min',
    '_min',
    '_min',
    '_min',
    '_min',
    '_min',
    '_next',
    '_next',
    '_oct',
    '_open',
    '_open',
    '_open',
    '_open',
    '_open',
    '_open',
    '_open',
    '_ord',
    '_print',
    '_print',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_pow',
    '_repr',
    '_round',
    '_round',
    '_setattr',
    '_sorted',
    '_sorted',
    '_sum',
    '_sum',
    '_sum',
    '_vars',
    '_vars',
]

__aliases__ = ['_PositiveInteger', '_NegativeInteger', '_IntegerFormats', '_Opener']

__annos__ = []

__constants__ = ['EnvironmentError', 'IOError']

__all__ = [
    '_object',
    '_staticmethod',
    '_classmethod',
    '_type',
    '_super',
    '_int',
    '_float',
    '_complex',
    '_FormatMapMapping',
    '_TranslateTable',
    '_str',
    '_bytes',
    '_bytearray',
    '_memoryview',
    '_bool',
    '_slice',
    '_tuple',
    '_list',
    '_dict',
    '_set',
    '_frozenset',
    '_enumerate',
    '_range',
    '_property',
    '_NotImplementedType',
    '_filter',
    '_GetItemIterable',
    '_map',
    '_SupportsWriteAndFlush',
    '_SupportsPow2',
    '_SupportsPow3NoneOnly',
    '_SupportsPow3',
    '_reversed',
    '_SupportsRound1',
    '_SupportsRound2',
    '_SupportsSumWithNoDefaultGiven',
    '_BaseException',
    '_GeneratorExit',
    '_KeyboardInterrupt',
    '_SystemExit',
    '_Exception',
    '_StopIteration',
    '_StopAsyncIteration',
    '_abs',
    '_all',
    '_any',
    '_ascii',
    '_bin',
    '_breakpoint',
    '_callable',
    '_chr',
    '_compile',
    '_delattr',
    '_dir',
    '_divmod',
    '_format',
    '_getattr',
    '_globals',
    '_hasattr',
    '_hash',
    '_hex',
    '_id',
    '_input',
    '_iter',
    '_isinstance',
    '_issubclass',
    '_len',
    '_locals',
    '_max',
    '_min',
    '_next',
    '_oct',
    '_open',
    '_ord',
    '_print',
    '_pow',
    '_repr',
    '_round',
    '_setattr',
    '_sorted',
    '_sum',
    '_vars',
]

################################################################################


class _object:
    __doc__: str | None
    __dict__: dict[str, Any]
    __module__: str
    __annotations__: dict[str, Any]

    @property
    def __class__(self) -> type: ...
    @__class__.setter
    def __class__(self, type: type, /) -> None: ...
    def __init__(self) -> None: ...
    def __new__(cls) -> Any: ...
    def __setattr__(self, name: str, value: Any, /) -> None: ...
    def __delattr__(self, name: str, /) -> None: ...

    def __lt__(self, value: object, /) -> bool: ...
    def __le__(self, value: object, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __gt__(self, value: object, /) -> bool: ...
    def __ge__(self, value: object, /) -> bool: ...

    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __format__(self, format_spec: str, /) -> str: ...
    def __getattribute__(self, name: str, /) -> Any: ...
    def __sizeof__(self) -> int: ...
    def __reduce__(self) -> str | tuple: ...
    def __reduce_ex__(self, protocol: int, /) -> str | tuple: ...

    if sys.version_info >= (3, 11):

        def __getstate__(self) -> object: ...

    def __dir__(self) -> Iterable[str]: ...
    def __init_subclass__(cls) -> None: ...
    @classmethod
    def __subclasshook__(cls, subclass: type, /) -> bool: ...


################################################################################


class _staticmethod:
    @property
    def __func__(self) -> Callable: ...
    @property
    def __isabstractmethod__(self) -> bool: ...
    def __init__(self, f: Callable, /) -> None: ...
    @overload
    def __get__(self, instance: None, owner: type, /) -> Callable: ...
    @overload
    def __get__(self, instance: Any, owner: type | None = None, /) -> Callable: ...

    if sys.version_info >= (3, 10):
        __name__: str
        __qualname__: str

        @property
        def __wrapped__(self) -> Callable: ...
        def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    if sys.version_info >= (3, 14):

        def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...

        __annotate__: AnnotateFunc | None


################################################################################


class _classmethod:
    @property
    def __func__(self) -> Callable: ...
    @property
    def __isabstractmethod__(self) -> bool: ...
    def __init__(self, f: Callable, /) -> None: ...
    @overload
    def __get__(self, instance: Any, owner: type | None = None, /) -> Callable: ...
    @overload
    def __get__(self, instance: None, owner: type, /) -> Callable: ...

    if sys.version_info >= (3, 10):
        __name__: str
        __qualname__: str

        @property
        def __wrapped__(self) -> Callable: ...

    if sys.version_info >= (3, 14):

        def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...

        __annotate__: AnnotateFunc | None


################################################################################


class _type:
    __bases__: tuple[type, ...]
    __module__: str
    __name__: str
    __qualname__: str
    __annotations__: dict[str, Any]

    @property
    def __base__(self) -> type | None: ...
    @property
    def __basicsize__(self) -> int: ...
    @property
    def __dict__(self) -> types.MappingProxyType[str, Any]: ...
    @property
    def __dictoffset__(self) -> int: ...
    @property
    def __flags__(self) -> int: ...
    @property
    def __itemsize__(self) -> int: ...
    @property
    def __mro__(self) -> tuple[type, ...]: ...
    @property
    def __text_signature__(self) -> str | None: ...
    @property
    def __weakrefoffset__(self) -> int: ...
    @overload
    def __init__(self, o: object, /) -> None: ...
    @overload
    def __init__(
        self, name: str, bases: tuple[type, ...], dict: dict[str, Any], /, **kwds: Any
    ) -> None: ...
    @overload
    def __new__(cls, o: object, /) -> type: ...
    @overload
    def __new__(
        cls: type, name: str, bases: tuple[type, ...], namespace: dict[str, Any], /, **kwds: Any
    ) -> Any: ...
    def __call__(self, *args: Any, **kwds: Any) -> Any: ...
    def __subclasses__(self: Any) -> list: ...
    def mro(self) -> list[type]: ...
    def __instancecheck__(self, instance: Any, /) -> bool: ...
    def __subclasscheck__(self, subclass: type, /) -> bool: ...
    @classmethod
    def __prepare__(
        metacls, name: str, bases: tuple[type, ...], /, **kwds: Any
    ) -> MutableMapping[str, object]: ...

    if sys.version_info >= (3, 10):

        def __or__(self, value: Any, /) -> types.UnionType: ...
        def __ror__(self, value: Any, /) -> types.UnionType: ...

    if sys.version_info >= (3, 12):
        __type_params__: tuple[TypeVar | Any | TypeVarTuple, ...]
    if sys.version_info >= (3, 14):
        __annotate__: AnnotateFunc | None


################################################################################


class _super:
    @overload
    def __init__(self, t: Any, obj: Any, /) -> None: ...
    @overload
    def __init__(self, t: Any, /) -> None: ...
    @overload
    def __init__(self) -> None: ...


################################################################################


class _int:
    @overload
    def __new__(cls, x: Any = ..., /) -> Any: ...
    @overload
    def __new__(cls, x: str | bytes | bytearray, /, base: int) -> Any: ...
    def as_integer_ratio(self) -> tuple[int, Literal[1]]: ...
    @property
    def real(self) -> int: ...
    @property
    def imag(self) -> Literal[0]: ...
    @property
    def numerator(self) -> int: ...
    @property
    def denominator(self) -> Literal[1]: ...
    def conjugate(self) -> int: ...
    def bit_length(self) -> int: ...

    if sys.version_info >= (3, 10):

        def bit_count(self) -> int: ...

    if sys.version_info >= (3, 11):

        def to_bytes(
            self,
            length: int = 1,
            byteorder: Literal["little", "big"] = "big",
            *,
            signed: bool = False,
        ) -> bytes: ...
        @classmethod
        def from_bytes(
            cls,
            bytes: Iterable[int] | SupportsBytes | Buffer,
            byteorder: Literal["little", "big"] = "big",
            *,
            signed: bool = False,
        ) -> Any: ...
    else:

        def to_bytes(
            self, length: int, byteorder: Literal["little", "big"], *, signed: bool = False
        ) -> bytes: ...
        @classmethod
        def from_bytes(
            cls,
            bytes: Iterable[int] | SupportsBytes | Buffer,
            byteorder: Literal["little", "big"],
            *,
            signed: bool = False,
        ) -> Any: ...

    if sys.version_info >= (3, 12):

        def is_integer(self) -> Literal[True]: ...

    def __add__(self, value: int, /) -> int: ...
    def __sub__(self, value: int, /) -> int: ...
    def __mul__(self, value: int, /) -> int: ...
    def __floordiv__(self, value: int, /) -> int: ...
    def __truediv__(self, value: int, /) -> float: ...
    def __mod__(self, value: int, /) -> int: ...
    def __divmod__(self, value: int, /) -> tuple[int, int]: ...
    def __radd__(self, value: int, /) -> int: ...
    def __rsub__(self, value: int, /) -> int: ...
    def __rmul__(self, value: int, /) -> int: ...
    def __rfloordiv__(self, value: int, /) -> int: ...
    def __rtruediv__(self, value: int, /) -> float: ...
    def __rmod__(self, value: int, /) -> int: ...
    def __rdivmod__(self, value: int, /) -> tuple[int, int]: ...
    @overload
    def __pow__(self, x: Literal[0], /) -> Literal[1]: ...
    @overload
    def __pow__(self, value: Literal[0], mod: None, /) -> Literal[1]: ...
    @overload
    def __pow__(
        self,
        value: Literal[
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            23,
            24,
            25,
        ],
        mod: None = None,
        /,
    ) -> int: ...
    @overload
    def __pow__(
        self,
        value: Literal[
            -1,
            -2,
            -3,
            -4,
            -5,
            -6,
            -7,
            -8,
            -9,
            -10,
            -11,
            -12,
            -13,
            -14,
            -15,
            -16,
            -17,
            -18,
            -19,
            -20,
        ],
        mod: None = None,
        /,
    ) -> float: ...
    @overload
    def __pow__(self, value: int, mod: None = None, /) -> Any: ...
    @overload
    def __pow__(self, value: int, mod: int, /) -> int: ...
    def __rpow__(self, value: int, mod: int | None = None, /) -> Any: ...
    def __and__(self, value: int, /) -> int: ...
    def __or__(self, value: int, /) -> int: ...
    def __xor__(self, value: int, /) -> int: ...
    def __lshift__(self, value: int, /) -> int: ...
    def __rshift__(self, value: int, /) -> int: ...
    def __rand__(self, value: int, /) -> int: ...
    def __ror__(self, value: int, /) -> int: ...
    def __rxor__(self, value: int, /) -> int: ...
    def __rlshift__(self, value: int, /) -> int: ...
    def __rrshift__(self, value: int, /) -> int: ...
    def __neg__(self) -> int: ...
    def __pos__(self) -> int: ...
    def __invert__(self) -> int: ...
    def __trunc__(self) -> int: ...
    def __ceil__(self) -> int: ...
    def __floor__(self) -> int: ...

    if sys.version_info >= (3, 14):

        def __round__(self, ndigits: int | None = None, /) -> int: ...
    else:

        def __round__(self, ndigits: int = ..., /) -> int: ...

    def __getnewargs__(self) -> tuple[int]: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __lt__(self, value: int, /) -> bool: ...
    def __le__(self, value: int, /) -> bool: ...
    def __gt__(self, value: int, /) -> bool: ...
    def __ge__(self, value: int, /) -> bool: ...
    def __float__(self) -> float: ...
    def __int__(self) -> int: ...
    def __abs__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __bool__(self) -> bool: ...
    def __index__(self) -> int: ...


################################################################################


class _float:
    def __new__(cls, x: str | Buffer | float | int = ..., /) -> Any: ...
    def as_integer_ratio(self) -> tuple[int, int]: ...
    def hex(self) -> str: ...
    def is_integer(self) -> bool: ...
    @classmethod
    def fromhex(cls, string: str, /) -> Any: ...
    @property
    def real(self) -> float: ...
    @property
    def imag(self) -> float: ...
    def conjugate(self) -> float: ...
    def __add__(self, value: float, /) -> float: ...
    def __sub__(self, value: float, /) -> float: ...
    def __mul__(self, value: float, /) -> float: ...
    def __floordiv__(self, value: float, /) -> float: ...
    def __truediv__(self, value: float, /) -> float: ...
    def __mod__(self, value: float, /) -> float: ...
    def __divmod__(self, value: float, /) -> tuple[float, float]: ...
    @overload
    def __pow__(self, value: int, mod: None = None, /) -> float: ...
    @overload
    def __pow__(self, value: float, mod: None = None, /) -> Any: ...
    def __radd__(self, value: float, /) -> float: ...
    def __rsub__(self, value: float, /) -> float: ...
    def __rmul__(self, value: float, /) -> float: ...
    def __rfloordiv__(self, value: float, /) -> float: ...
    def __rtruediv__(self, value: float, /) -> float: ...
    def __rmod__(self, value: float, /) -> float: ...
    def __rdivmod__(self, value: float, /) -> tuple[float, float]: ...
    @overload
    def __rpow__(
        self,
        value: Literal[
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            23,
            24,
            25,
        ],
        mod: None = None,
        /,
    ) -> float: ...
    @overload
    def __rpow__(
        self,
        value: Literal[
            -1,
            -2,
            -3,
            -4,
            -5,
            -6,
            -7,
            -8,
            -9,
            -10,
            -11,
            -12,
            -13,
            -14,
            -15,
            -16,
            -17,
            -18,
            -19,
            -20,
        ],
        mod: None = None,
        /,
    ) -> complex: ...
    @overload
    def __rpow__(self, value: float, mod: None = None, /) -> Any: ...
    def __getnewargs__(self) -> tuple[float]: ...
    def __trunc__(self) -> int: ...
    def __ceil__(self) -> int: ...
    def __floor__(self) -> int: ...
    @overload
    def __round__(self, ndigits: None = None, /) -> int: ...
    @overload
    def __round__(self, ndigits: int, /) -> float: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __lt__(self, value: float, /) -> bool: ...
    def __le__(self, value: float, /) -> bool: ...
    def __gt__(self, value: float, /) -> bool: ...
    def __ge__(self, value: float, /) -> bool: ...
    def __neg__(self) -> float: ...
    def __pos__(self) -> float: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __abs__(self) -> float: ...
    def __hash__(self) -> int: ...
    def __bool__(self) -> bool: ...

    if sys.version_info >= (3, 14):

        @classmethod
        def from_number(cls, number: float | int | float, /) -> Any: ...


################################################################################


class _complex:
    @overload
    def __new__(
        cls, real: complex | SupportsComplex | float | int = ..., imag: complex | float | int = ...
    ) -> Any: ...
    @overload
    def __new__(cls, real: str | SupportsComplex | float | int | complex) -> Any: ...
    @property
    def real(self) -> float: ...
    @property
    def imag(self) -> float: ...
    def conjugate(self) -> complex: ...
    def __add__(self, value: complex, /) -> complex: ...
    def __sub__(self, value: complex, /) -> complex: ...
    def __mul__(self, value: complex, /) -> complex: ...
    def __pow__(self, value: complex, mod: None = None, /) -> complex: ...
    def __truediv__(self, value: complex, /) -> complex: ...
    def __radd__(self, value: complex, /) -> complex: ...
    def __rsub__(self, value: complex, /) -> complex: ...
    def __rmul__(self, value: complex, /) -> complex: ...
    def __rpow__(self, value: complex, mod: None = None, /) -> complex: ...
    def __rtruediv__(self, value: complex, /) -> complex: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __neg__(self) -> complex: ...
    def __pos__(self) -> complex: ...
    def __abs__(self) -> float: ...
    def __hash__(self) -> int: ...
    def __bool__(self) -> bool: ...

    if sys.version_info >= (3, 11):

        def __complex__(self) -> complex: ...

    if sys.version_info >= (3, 14):

        @classmethod
        def from_number(cls, number: complex | SupportsComplex | float | int, /) -> Any: ...


################################################################################


class _FormatMapMapping:
    def __getitem__(self, key: str, /) -> Any: ...


################################################################################


class _TranslateTable:
    def __getitem__(self, key: int, /) -> str | int | None: ...


################################################################################


class _str:
    @overload
    def __new__(cls, object: object = ...) -> Any: ...
    @overload
    def __new__(cls, object: Buffer, encoding: str = ..., errors: str = ...) -> Any: ...
    @overload
    def capitalize(self) -> str: ...
    @overload
    def casefold(self) -> str: ...
    @overload
    def center(self, width: int, fillchar: str = " ", /) -> str: ...
    def count(self, sub: str, start: int | None = ..., end: int | None = ..., /) -> int: ...
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes: ...
    def endswith(
        self, suffix: str | tuple[str, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    @overload
    def expandtabs(self, tabsize: int = 8) -> str: ...
    def find(self, sub: str, start: int | None = ..., end: int | None = ..., /) -> int: ...
    @overload
    def format(self, *args: object, **kwargs: object) -> str: ...
    def format_map(self, mapping: _FormatMapMapping, /) -> str: ...
    def index(self, sub: str, start: int | None = ..., end: int | None = ..., /) -> int: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    def isascii(self) -> bool: ...
    def isdecimal(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def isidentifier(self) -> bool: ...
    def islower(self) -> bool: ...
    def isnumeric(self) -> bool: ...
    def isprintable(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    @overload
    def join(self, iterable: Iterable[str], /) -> str: ...
    @overload
    def ljust(self, width: int, fillchar: str = " ", /) -> str: ...
    @overload
    def lower(self) -> str: ...
    @overload
    def lstrip(self, chars: str | None = None, /) -> str: ...
    @overload
    def partition(self, sep: str, /) -> tuple[str, str, str]: ...

    if sys.version_info >= (3, 13):

        @overload
        def replace(self, old: str, new: str, /, count: int = -1) -> str: ...
    else:

        @overload
        def replace(self, old: str, new: str, count: int = -1, /) -> str: ...

    @overload
    def removeprefix(self, prefix: str, /) -> str: ...
    @overload
    def removesuffix(self, suffix: str, /) -> str: ...
    def rfind(self, sub: str, start: int | None = ..., end: int | None = ..., /) -> int: ...
    def rindex(self, sub: str, start: int | None = ..., end: int | None = ..., /) -> int: ...
    @overload
    def rjust(self, width: int, fillchar: str = " ", /) -> str: ...
    @overload
    def rpartition(self, sep: str, /) -> tuple[str, str, str]: ...
    @overload
    def rsplit(self, sep: str | None = None, maxsplit: int = -1) -> list[str]: ...
    @overload
    def rstrip(self, chars: str | None = None, /) -> str: ...
    @overload
    def split(self, sep: str | None = None, maxsplit: int = -1) -> list[str]: ...
    @overload
    def splitlines(self, keepends: bool = False) -> list[str]: ...
    def startswith(
        self, prefix: str | tuple[str, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    @overload
    def strip(self, chars: str | None = None, /) -> str: ...
    @overload
    def swapcase(self) -> str: ...
    @overload
    def title(self) -> str: ...
    def translate(self, table: _TranslateTable, /) -> str: ...
    @overload
    def upper(self) -> str: ...
    @overload
    def zfill(self, width: int, /) -> str: ...
    @staticmethod
    @overload
    def maketrans(
        x: dict[int, Any] | dict[str, Any] | dict[str | int, Any], /
    ) -> dict[int, Any]: ...
    @staticmethod
    @overload
    def maketrans(x: str, y: str, /) -> dict[int, int]: ...
    @staticmethod
    @overload
    def maketrans(x: str, y: str, z: str, /) -> dict[int, int | None]: ...
    @overload
    def __add__(self, value: str, /) -> str: ...
    def __contains__(self, key: str, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ge__(self, value: str, /) -> bool: ...
    @overload
    def __getitem__(self, key: int | slice, /) -> str: ...
    def __gt__(self, value: str, /) -> bool: ...
    def __hash__(self) -> int: ...
    @overload
    def __iter__(self) -> Iterator[str]: ...
    def __le__(self, value: str, /) -> bool: ...
    def __len__(self) -> int: ...
    def __lt__(self, value: str, /) -> bool: ...
    @overload
    def __mod__(self, value: Any, /) -> str: ...
    @overload
    def __mul__(self, value: int, /) -> str: ...
    def __ne__(self, value: object, /) -> bool: ...
    @overload
    def __rmul__(self, value: int, /) -> str: ...
    def __getnewargs__(self) -> tuple[str]: ...


################################################################################


class _bytes:
    @overload
    def __new__(cls, o: Iterable[int] | int | SupportsBytes | Buffer, /) -> Any: ...
    @overload
    def __new__(cls, string: str, /, encoding: str, errors: str = ...) -> Any: ...
    @overload
    def __new__(cls) -> Any: ...
    def capitalize(self) -> bytes: ...
    def center(self, width: int, fillchar: bytes = b" ", /) -> bytes: ...
    def count(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def decode(self, encoding: str = "utf-8", errors: str = "strict") -> str: ...
    def endswith(
        self, suffix: Buffer | tuple[Buffer, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    def expandtabs(self, tabsize: int = 8) -> bytes: ...
    def find(self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /) -> int: ...
    def hex(self, sep: str | bytes = ..., bytes_per_sep: int = ...) -> str: ...
    def index(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    def isascii(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def islower(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    def join(self, iterable_of_bytes: Iterable[Buffer], /) -> bytes: ...
    def ljust(self, width: int, fillchar: bytes | bytearray = b" ", /) -> bytes: ...
    def lower(self) -> bytes: ...
    def lstrip(self, bytes: Buffer | None = None, /) -> bytes: ...
    def partition(self, sep: Buffer, /) -> tuple[bytes, bytes, bytes]: ...
    def replace(self, old: Buffer, new: Buffer, count: int = -1, /) -> bytes: ...
    def removeprefix(self, prefix: Buffer, /) -> bytes: ...
    def removesuffix(self, suffix: Buffer, /) -> bytes: ...
    def rfind(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def rindex(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def rjust(self, width: int, fillchar: bytes | bytearray = b" ", /) -> bytes: ...
    def rpartition(self, sep: Buffer, /) -> tuple[bytes, bytes, bytes]: ...
    def rsplit(self, sep: Buffer | None = None, maxsplit: int = -1) -> list[bytes]: ...
    def rstrip(self, bytes: Buffer | None = None, /) -> bytes: ...
    def split(self, sep: Buffer | None = None, maxsplit: int = -1) -> list[bytes]: ...
    def splitlines(self, keepends: bool = False) -> list[bytes]: ...
    def startswith(
        self, prefix: Buffer | tuple[Buffer, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    def strip(self, bytes: Buffer | None = None, /) -> bytes: ...
    def swapcase(self) -> bytes: ...
    def title(self) -> bytes: ...
    def translate(self, table: Buffer | None, /, delete: bytes = b"") -> bytes: ...
    def upper(self) -> bytes: ...
    def zfill(self, width: int, /) -> bytes: ...
    @classmethod
    def fromhex(cls, string: str, /) -> Any: ...
    @staticmethod
    def maketrans(frm: Buffer, to: Buffer, /) -> bytes: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[int]: ...
    def __hash__(self) -> int: ...
    @overload
    def __getitem__(self, key: int, /) -> int: ...
    @overload
    def __getitem__(self, key: slice, /) -> bytes: ...
    def __add__(self, value: Buffer, /) -> bytes: ...
    def __mul__(self, value: int, /) -> bytes: ...
    def __rmul__(self, value: int, /) -> bytes: ...
    def __mod__(self, value: Any, /) -> bytes: ...
    def __contains__(self, key: int | Buffer, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __lt__(self, value: bytes, /) -> bool: ...
    def __le__(self, value: bytes, /) -> bool: ...
    def __gt__(self, value: bytes, /) -> bool: ...
    def __ge__(self, value: bytes, /) -> bool: ...
    def __getnewargs__(self) -> tuple[bytes]: ...

    if sys.version_info >= (3, 11):

        def __bytes__(self) -> bytes: ...

    def __buffer__(self, flags: int, /) -> memoryview: ...


################################################################################


class _bytearray:
    __hash__: ClassVar[None]

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, ints: Iterable[int] | int | Buffer, /) -> None: ...
    @overload
    def __init__(self, string: str, /, encoding: str, errors: str = ...) -> None: ...
    def append(self, item: int, /) -> None: ...
    def capitalize(self) -> bytearray: ...
    def center(self, width: int, fillchar: bytes = b" ", /) -> bytearray: ...
    def count(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def copy(self) -> bytearray: ...
    def decode(self, encoding: str = "utf-8", errors: str = "strict") -> str: ...
    def endswith(
        self, suffix: Buffer | tuple[Buffer, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    def expandtabs(self, tabsize: int = 8) -> bytearray: ...
    def extend(self, iterable_of_ints: Iterable[int], /) -> None: ...
    def find(self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /) -> int: ...
    def hex(self, sep: str | bytes = ..., bytes_per_sep: int = ...) -> str: ...
    def index(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def insert(self, index: int, item: int, /) -> None: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    def isascii(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def islower(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    def join(self, iterable_of_bytes: Iterable[Buffer], /) -> bytearray: ...
    def ljust(self, width: int, fillchar: bytes | bytearray = b" ", /) -> bytearray: ...
    def lower(self) -> bytearray: ...
    def lstrip(self, bytes: Buffer | None = None, /) -> bytearray: ...
    def partition(self, sep: Buffer, /) -> tuple[bytearray, bytearray, bytearray]: ...
    def pop(self, index: int = -1, /) -> int: ...
    def remove(self, value: int, /) -> None: ...
    def removeprefix(self, prefix: Buffer, /) -> bytearray: ...
    def removesuffix(self, suffix: Buffer, /) -> bytearray: ...
    def replace(self, old: Buffer, new: Buffer, count: int = -1, /) -> bytearray: ...
    def rfind(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def rindex(
        self, sub: Buffer | int, start: int | None = ..., end: int | None = ..., /
    ) -> int: ...
    def rjust(self, width: int, fillchar: bytes | bytearray = b" ", /) -> bytearray: ...
    def rpartition(self, sep: Buffer, /) -> tuple[bytearray, bytearray, bytearray]: ...
    def rsplit(self, sep: Buffer | None = None, maxsplit: int = -1) -> list[bytearray]: ...
    def rstrip(self, bytes: Buffer | None = None, /) -> bytearray: ...
    def split(self, sep: Buffer | None = None, maxsplit: int = -1) -> list[bytearray]: ...
    def splitlines(self, keepends: bool = False) -> list[bytearray]: ...
    def startswith(
        self, prefix: Buffer | tuple[Buffer, ...], start: int | None = ..., end: int | None = ..., /
    ) -> bool: ...
    def strip(self, bytes: Buffer | None = None, /) -> bytearray: ...
    def swapcase(self) -> bytearray: ...
    def title(self) -> bytearray: ...
    def translate(self, table: Buffer | None, /, delete: bytes = b"") -> bytearray: ...
    def upper(self) -> bytearray: ...
    def zfill(self, width: int, /) -> bytearray: ...
    @classmethod
    def fromhex(cls, string: str, /) -> Any: ...
    @staticmethod
    def maketrans(frm: Buffer, to: Buffer, /) -> bytes: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[int]: ...
    @overload
    def __getitem__(self, key: int, /) -> int: ...
    @overload
    def __getitem__(self, key: slice, /) -> bytearray: ...
    @overload
    def __setitem__(self, key: int, value: int, /) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Iterable[int] | bytes, /) -> None: ...
    def __delitem__(self, key: int | slice, /) -> None: ...
    def __add__(self, value: Buffer, /) -> bytearray: ...
    def __iadd__(self, value: Buffer, /) -> Any: ...
    def __mul__(self, value: int, /) -> bytearray: ...
    def __rmul__(self, value: int, /) -> bytearray: ...
    def __imul__(self, value: int, /) -> Any: ...
    def __mod__(self, value: Any, /) -> bytes: ...
    def __contains__(self, key: int | Buffer, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __ne__(self, value: object, /) -> bool: ...
    def __lt__(self, value: Buffer, /) -> bool: ...
    def __le__(self, value: Buffer, /) -> bool: ...
    def __gt__(self, value: Buffer, /) -> bool: ...
    def __ge__(self, value: Buffer, /) -> bool: ...
    def __alloc__(self) -> int: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...
    def __release_buffer__(self, buffer: memoryview, /) -> None: ...

    if sys.version_info >= (3, 14):

        def resize(self, size: int, /) -> None: ...


################################################################################


class _memoryview:
    index: ClassVar[None]
    count: ClassVar[None]

    @property
    def format(self) -> str: ...
    @property
    def itemsize(self) -> int: ...
    @property
    def shape(self) -> tuple[int, ...] | None: ...
    @property
    def strides(self) -> tuple[int, ...] | None: ...
    @property
    def suboffsets(self) -> tuple[int, ...] | None: ...
    @property
    def readonly(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def obj(self) -> Buffer: ...
    @property
    def c_contiguous(self) -> bool: ...
    @property
    def f_contiguous(self) -> bool: ...
    @property
    def contiguous(self) -> bool: ...
    @property
    def nbytes(self) -> int: ...
    def __new__(cls, obj: Buffer) -> Any: ...
    def __enter__(self) -> Any: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
        /,
    ) -> None: ...
    @overload
    def cast(
        self, format: Literal["c", "@c"], shape: list[int] | tuple[int, ...] = ...
    ) -> memoryview: ...
    @overload
    def cast(
        self, format: Literal["f", "@f", "d", "@d"], shape: list[int] | tuple[int, ...] = ...
    ) -> memoryview: ...
    @overload
    def cast(
        self, format: Literal["?"], shape: list[int] | tuple[int, ...] = ...
    ) -> memoryview: ...
    @overload
    def cast(
        self,
        format: Literal[
            "b",
            "B",
            "@b",
            "@B",
            "h",
            "H",
            "@h",
            "@H",
            "i",
            "I",
            "@i",
            "@I",
            "l",
            "L",
            "@l",
            "@L",
            "q",
            "Q",
            "@q",
            "@Q",
            "P",
            "@P",
        ],
        shape: list[int] | tuple[int, ...] = ...,
    ) -> memoryview: ...
    @overload
    def __getitem__(self, key: int | tuple[int, ...], /) -> Any: ...
    @overload
    def __getitem__(self, key: slice, /) -> memoryview: ...
    def __contains__(self, x: object, /) -> bool: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    @overload
    def __setitem__(self, key: slice, value: Buffer, /) -> None: ...
    @overload
    def __setitem__(self, key: int | tuple[int, ...], value: Any, /) -> None: ...

    if sys.version_info >= (3, 10):

        def tobytes(self, order: Literal["C", "F", "A"] | None = "C") -> bytes: ...
    else:

        def tobytes(self, order: Literal["C", "F", "A"] | None = None) -> bytes: ...

    def tolist(self) -> list[int]: ...
    def toreadonly(self) -> memoryview: ...
    def release(self) -> None: ...
    def hex(self, sep: str | bytes = ..., bytes_per_sep: int = ...) -> str: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...
    def __release_buffer__(self, buffer: memoryview, /) -> None: ...

    if sys.version_info >= (3, 14):

        def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class _bool:
    def __new__(cls, o: object = ..., /) -> Any: ...
    @overload
    def __and__(self, value: bool, /) -> bool: ...
    @overload
    def __and__(self, value: int, /) -> int: ...
    @overload
    def __or__(self, value: bool, /) -> bool: ...
    @overload
    def __or__(self, value: int, /) -> int: ...
    @overload
    def __xor__(self, value: bool, /) -> bool: ...
    @overload
    def __xor__(self, value: int, /) -> int: ...
    @overload
    def __rand__(self, value: bool, /) -> bool: ...
    @overload
    def __rand__(self, value: int, /) -> int: ...
    @overload
    def __ror__(self, value: bool, /) -> bool: ...
    @overload
    def __ror__(self, value: int, /) -> int: ...
    @overload
    def __rxor__(self, value: bool, /) -> bool: ...
    @overload
    def __rxor__(self, value: int, /) -> int: ...
    def __getnewargs__(self) -> tuple[int]: ...
    def __invert__(self) -> int: ...


################################################################################


class _slice:
    @property
    def start(self) -> Any: ...
    @property
    def step(self) -> Any: ...
    @property
    def stop(self) -> Any: ...
    @overload
    def __new__(cls, start: None, stop: None = None, step: None = None, /) -> slice: ...
    @overload
    def __new__(cls, stop: Any, /) -> slice: ...
    @overload
    def __new__(cls, start: Any, stop: None, step: None = None, /) -> slice: ...
    @overload
    def __new__(cls, start: None, stop: Any, step: None = None, /) -> slice: ...
    @overload
    def __new__(cls, start: Any, stop: Any, step: None = None, /) -> slice: ...
    @overload
    def __new__(cls, start: None, stop: None, step: Any, /) -> slice: ...
    @overload
    def __new__(cls, start: Any, stop: None, step: Any, /) -> slice: ...
    @overload
    def __new__(cls, start: None, stop: Any, step: Any, /) -> slice: ...
    @overload
    def __new__(cls, start: Any, stop: Any, step: Any, /) -> slice: ...
    def __eq__(self, value: object, /) -> bool: ...

    if sys.version_info >= (3, 12):

        def __hash__(self) -> int: ...
    else:
        __hash__: ClassVar[None]

    def indices(self, len: int, /) -> tuple[int, int, int]: ...


################################################################################


class _tuple:
    def __new__(cls, iterable: Iterable = ..., /) -> Any: ...
    def __len__(self) -> int: ...
    def __contains__(self, key: object, /) -> bool: ...
    @overload
    def __getitem__(self, key: int, /) -> Any: ...
    @overload
    def __getitem__(self, key: slice, /) -> tuple: ...
    def __iter__(self) -> Iterator: ...
    def __lt__(self, value: tuple, /) -> bool: ...
    def __le__(self, value: tuple, /) -> bool: ...
    def __gt__(self, value: tuple, /) -> bool: ...
    def __ge__(self, value: tuple, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    @overload
    def __add__(self, value: tuple, /) -> tuple: ...
    @overload
    def __add__(self, value: tuple, /) -> tuple: ...
    def __mul__(self, value: int, /) -> tuple: ...
    def __rmul__(self, value: int, /) -> tuple: ...
    def count(self, value: Any, /) -> int: ...
    def index(self, value: Any, start: int = 0, stop: int = sys.maxsize, /) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class FunctionType:
    __code__: CodeType
    __defaults__: tuple | None
    __dict__: dict[str, Any]
    __name__: str
    __qualname__: str
    __annotations__: dict[str, Any]
    __kwdefaults__: dict[str, Any] | None
    __module__: str

    @property
    def __closure__(self) -> tuple | None: ...
    @property
    def __globals__(self) -> dict[str, Any]: ...

    if sys.version_info >= (3, 14):
        __annotate__: AnnotateFunc | None
    if sys.version_info >= (3, 10):

        @property
        def __builtins__(self) -> dict[str, Any]: ...

    if sys.version_info >= (3, 12):
        __type_params__: tuple
    if sys.version_info >= (3, 13):

        def __new__(
            cls,
            code: CodeType,
            globals: dict[str, Any],
            name: str | None = None,
            argdefs: tuple | None = None,
            closure: tuple | None = None,
            kwdefaults: dict[str, object] | None = None,
        ) -> Any: ...
    else:

        def __new__(
            cls,
            code: CodeType,
            globals: dict[str, Any],
            name: str | None = None,
            argdefs: tuple | None = None,
            closure: tuple | None = None,
        ) -> Any: ...

    def __get__(self, instance: object, owner: type | None = None, /) -> Any: ...


################################################################################


class _list:
    __hash__: ClassVar[None]

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, iterable: Iterable, /) -> None: ...
    def copy(self) -> list: ...
    def append(self, object: Any, /) -> None: ...
    def extend(self, iterable: Iterable, /) -> None: ...
    def pop(self, index: int = -1, /) -> Any: ...
    def index(self, value: Any, start: int = 0, stop: int = sys.maxsize, /) -> int: ...
    def count(self, value: Any, /) -> int: ...
    def insert(self, index: int, object: Any, /) -> None: ...
    def remove(self, value: Any, /) -> None: ...
    @overload
    def sort(self: list, *, key: None = None, reverse: bool = False) -> None: ...
    @overload
    def sort(
        self, *, key: Callable[[Any], SupportsDunderLT | SupportsDunderGT], reverse: bool = False
    ) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator: ...
    @overload
    def __getitem__(self, i: int, /) -> Any: ...
    @overload
    def __getitem__(self, s: slice, /) -> list: ...
    @overload
    def __setitem__(self, key: int, value: Any, /) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Iterable, /) -> None: ...
    def __delitem__(self, key: int | slice, /) -> None: ...
    @overload
    def __add__(self, value: list, /) -> list: ...
    @overload
    def __add__(self, value: list, /) -> list: ...
    def __iadd__(self, value: Iterable, /) -> Any: ...
    def __mul__(self, value: int, /) -> list: ...
    def __rmul__(self, value: int, /) -> list: ...
    def __imul__(self, value: int, /) -> Any: ...
    def __contains__(self, key: object, /) -> bool: ...
    def __reversed__(self) -> Iterator: ...
    def __gt__(self, value: list, /) -> bool: ...
    def __ge__(self, value: list, /) -> bool: ...
    def __lt__(self, value: list, /) -> bool: ...
    def __le__(self, value: list, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class _dict:
    __hash__: ClassVar[None]

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self: dict, **kwargs: Any) -> None: ...  # pyright: ignore[reportInvalidTypeVarUse]  #11780
    @overload
    def __init__(self, map: SupportsKeysAndGetItem, /) -> None: ...
    @overload
    def __init__(
        self: dict,  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        map: SupportsKeysAndGetItem,
        /,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(self, iterable: Iterable, /) -> None: ...
    @overload
    def __init__(
        self: dict,  # pyright: ignore[reportInvalidTypeVarUse]  #11780
        iterable: Iterable,
        /,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(self: dict, iterable: Iterable[list], /) -> None: ...
    @overload
    def __init__(self: dict, iterable: Iterable[list], /) -> None: ...
    def __new__(cls, *args: Any, **kwargs: Any) -> Any: ...
    def copy(self) -> dict: ...
    def keys(self) -> dict_keys: ...
    def values(self) -> dict_values: ...
    def items(self) -> dict_items: ...
    @classmethod
    @overload
    def fromkeys(cls, iterable: Iterable, value: None = None, /) -> dict: ...
    @classmethod
    @overload
    def fromkeys(cls, iterable: Iterable, value: Any, /) -> dict: ...
    @overload
    def get(self, key: Any, default: None = None, /) -> Any: ...
    @overload
    def get(self, key: Any, default: Any, /) -> Any: ...
    @overload
    def get(self, key: Any, default: Any, /) -> Any: ...
    @overload
    def pop(self, key: Any, /) -> Any: ...
    @overload
    def pop(self, key: Any, default: Any, /) -> Any: ...
    @overload
    def pop(self, key: Any, default: Any, /) -> Any: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: Any, /) -> Any: ...
    def __setitem__(self, key: Any, value: Any, /) -> None: ...
    def __delitem__(self, key: Any, /) -> None: ...
    def __iter__(self) -> Iterator: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __reversed__(self) -> Iterator: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...
    @overload
    def __or__(self, value: dict, /) -> dict: ...
    @overload
    def __or__(self, value: dict, /) -> dict: ...
    @overload
    def __ror__(self, value: dict, /) -> dict: ...
    @overload
    def __ror__(self, value: dict, /) -> dict: ...
    @overload
    def __ior__(self, value: SupportsKeysAndGetItem, /) -> Any: ...
    @overload
    def __ior__(self, value: Iterable, /) -> Any: ...


################################################################################


class _set:
    __hash__: ClassVar[None]

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, iterable: Iterable, /) -> None: ...
    def add(self, element: Any, /) -> None: ...
    def copy(self) -> set: ...
    def difference(self, *s: Iterable) -> set: ...
    def difference_update(self, *s: Iterable) -> None: ...
    def discard(self, element: Any, /) -> None: ...
    def intersection(self, *s: Iterable) -> set: ...
    def intersection_update(self, *s: Iterable) -> None: ...
    def isdisjoint(self, s: Iterable, /) -> bool: ...
    def issubset(self, s: Iterable, /) -> bool: ...
    def issuperset(self, s: Iterable, /) -> bool: ...
    def remove(self, element: Any, /) -> None: ...
    def symmetric_difference(self, s: Iterable, /) -> set: ...
    def symmetric_difference_update(self, s: Iterable, /) -> None: ...
    def union(self, *s: Iterable) -> set: ...
    def update(self, *s: Iterable) -> None: ...
    def __len__(self) -> int: ...
    def __contains__(self, o: object, /) -> bool: ...
    def __iter__(self) -> Iterator: ...
    def __and__(self, value: AbstractSet, /) -> set: ...
    def __iand__(self, value: AbstractSet, /) -> Any: ...
    def __or__(self, value: AbstractSet, /) -> set: ...
    def __ior__(self, value: AbstractSet, /) -> Any: ...
    def __sub__(self, value: AbstractSet, /) -> set: ...
    def __isub__(self, value: AbstractSet, /) -> Any: ...
    def __xor__(self, value: AbstractSet, /) -> set: ...
    def __ixor__(self, value: AbstractSet, /) -> Any: ...
    def __le__(self, value: AbstractSet, /) -> bool: ...
    def __lt__(self, value: AbstractSet, /) -> bool: ...
    def __ge__(self, value: AbstractSet, /) -> bool: ...
    def __gt__(self, value: AbstractSet, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class _frozenset:
    @overload
    def __new__(cls) -> Any: ...
    @overload
    def __new__(cls, iterable: Iterable, /) -> Any: ...
    def copy(self) -> frozenset: ...
    def difference(self, *s: Iterable) -> frozenset: ...
    def intersection(self, *s: Iterable) -> frozenset: ...
    def isdisjoint(self, s: Iterable, /) -> bool: ...
    def issubset(self, s: Iterable, /) -> bool: ...
    def issuperset(self, s: Iterable, /) -> bool: ...
    def symmetric_difference(self, s: Iterable, /) -> frozenset: ...
    def union(self, *s: Iterable) -> frozenset: ...
    def __len__(self) -> int: ...
    def __contains__(self, o: object, /) -> bool: ...
    def __iter__(self) -> Iterator: ...
    def __and__(self, value: AbstractSet, /) -> frozenset: ...
    def __or__(self, value: AbstractSet, /) -> frozenset: ...
    def __sub__(self, value: AbstractSet, /) -> frozenset: ...
    def __xor__(self, value: AbstractSet, /) -> frozenset: ...
    def __le__(self, value: AbstractSet, /) -> bool: ...
    def __lt__(self, value: AbstractSet, /) -> bool: ...
    def __ge__(self, value: AbstractSet, /) -> bool: ...
    def __gt__(self, value: AbstractSet, /) -> bool: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class _enumerate:
    def __new__(cls, iterable: Iterable, start: int = 0) -> Any: ...
    def __iter__(self) -> Any: ...
    def __next__(self) -> tuple: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias: ...


################################################################################


class _range:
    @property
    def start(self) -> int: ...
    @property
    def stop(self) -> int: ...
    @property
    def step(self) -> int: ...
    @overload
    def __new__(cls, stop: int, /) -> Any: ...
    @overload
    def __new__(cls, start: int, stop: int, step: int = ..., /) -> Any: ...
    def count(self, value: int, /) -> int: ...
    def index(self, value: int, /) -> int: ...
    def __len__(self) -> int: ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...
    def __contains__(self, key: object, /) -> bool: ...
    def __iter__(self) -> Iterator[int]: ...
    @overload
    def __getitem__(self, key: int, /) -> int: ...
    @overload
    def __getitem__(self, key: slice, /) -> range: ...
    def __reversed__(self) -> Iterator[int]: ...


################################################################################


class _property:
    fget: Callable[[Any], Any] | None
    fset: Callable[[Any, Any], None] | None
    fdel: Callable[[Any], None] | None
    __isabstractmethod__: bool
    if sys.version_info >= (3, 13):
        __name__: str

    def __init__(
        self,
        fget: Callable[[Any], Any] | None = ...,
        fset: Callable[[Any, Any], None] | None = ...,
        fdel: Callable[[Any], None] | None = ...,
        doc: str | None = ...,
    ) -> None: ...
    def getter(self, fget: Callable[[Any], Any], /) -> property: ...
    def setter(self, fset: Callable[[Any, Any], None], /) -> property: ...
    def deleter(self, fdel: Callable[[Any], None], /) -> property: ...
    @overload
    def __get__(self, instance: None, owner: type, /) -> Any: ...
    @overload
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...
    def __set__(self, instance: Any, value: Any, /) -> None: ...
    def __delete__(self, instance: Any, /) -> None: ...


################################################################################


class _NotImplementedType:
    __call__: None


################################################################################


class _filter:
    @overload
    def __new__(cls, function: None, iterable: Iterable, /) -> Any: ...
    @overload
    def __new__(cls, function: Callable[[Any], bool], iterable: Iterable, /) -> Any: ...
    @overload
    def __new__(cls, function: Callable[[Any], bool], iterable: Iterable, /) -> Any: ...
    @overload
    def __new__(cls, function: Callable[[Any], Any], iterable: Iterable, /) -> Any: ...
    def __iter__(self) -> Any: ...
    def __next__(self) -> Any: ...


################################################################################


class _GetItemIterable:
    def __getitem__(self, i: int, /) -> Any: ...


################################################################################


class _map:
    if sys.version_info >= (3, 14):

        @overload
        def __new__(
            cls, func: Callable[[Any], Any], iterable: Iterable, /, *, strict: bool = False
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            /,
            *,
            strict: bool = False,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            /,
            *,
            strict: bool = False,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            /,
            *,
            strict: bool = False,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            iter5: Iterable,
            /,
            *,
            strict: bool = False,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable,
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            iter5: Iterable,
            iter6: Iterable,
            /,
            *iterables: Iterable,
            strict: bool = False,
        ) -> Any: ...
    else:

        @overload
        def __new__(cls, func: Callable[[Any], Any], iterable: Iterable, /) -> Any: ...
        @overload
        def __new__(
            cls, func: Callable[[Any, Any], Any], iterable: Iterable, iter2: Iterable, /
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            /,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            /,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable[[Any, Any, Any, Any, Any], Any],
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            iter5: Iterable,
            /,
        ) -> Any: ...
        @overload
        def __new__(
            cls,
            func: Callable,
            iterable: Iterable,
            iter2: Iterable,
            iter3: Iterable,
            iter4: Iterable,
            iter5: Iterable,
            iter6: Iterable,
            /,
            *iterables: Iterable,
        ) -> Any: ...

    def __iter__(self) -> Any: ...
    def __next__(self) -> Any: ...


################################################################################


class _SupportsWriteAndFlush:
    pass


################################################################################


class _SupportsPow2:
    def __pow__(self, other: Any, /) -> Any: ...


################################################################################


class _SupportsPow3NoneOnly:
    def __pow__(self, other: Any, modulo: None = None, /) -> Any: ...


################################################################################


class _SupportsPow3:
    def __pow__(self, other: Any, modulo: Any, /) -> Any: ...


################################################################################


class _reversed:
    @overload
    def __new__(cls, sequence: Reversible, /) -> Iterator: ...
    @overload
    def __new__(cls, sequence: SupportsLenAndGetItem, /) -> Iterator: ...
    def __iter__(self) -> Any: ...
    def __next__(self) -> Any: ...
    def __length_hint__(self) -> int: ...


################################################################################


class _SupportsRound1:
    def __round__(self) -> Any: ...


################################################################################


class _SupportsRound2:
    def __round__(self, ndigits: int, /) -> Any: ...


################################################################################


class _SupportsSumWithNoDefaultGiven:
    pass


################################################################################


class _BaseException:
    args: tuple
    __cause__: BaseException | None
    __context__: BaseException | None
    __suppress_context__: bool
    __traceback__: TracebackType | None

    def __init__(self, *args: object) -> None: ...
    def __new__(cls, *args: Any, **kwds: Any) -> Any: ...
    def __setstate__(self, state: dict | None, /) -> None: ...
    def with_traceback(self, tb: TracebackType | None, /) -> Any: ...

    if sys.version_info >= (3, 11):
        # only present after add_note() is called
        __notes__: list

        def add_note(self, note: str, /) -> None: ...


################################################################################


class _GeneratorExit:
    pass


################################################################################


class _KeyboardInterrupt:
    pass


################################################################################


class _SystemExit:
    code: int


################################################################################


class _Exception:
    pass


################################################################################


class _StopIteration:
    value: Any


################################################################################


class _StopAsyncIteration:
    pass


################################################################################


def _abs(x: SupportsAbs, /) -> Any: ...


def _all(iterable: Iterable, /) -> bool: ...


def _any(iterable: Iterable, /) -> bool: ...


def _ascii(obj: object, /) -> str: ...


def _bin(number: int | int, /) -> str: ...


def _breakpoint(*args: Any, **kws: Any) -> None: ...


def _callable(obj: object, /) -> bool: ...


def _chr(i: int | int, /) -> str: ...


@overload
def _compile(
    source: str | Buffer | ast.AST | ast.AST | ast.AST,
    filename: str | Buffer | PathLike,
    mode: str,
    flags: Literal[0],
    dont_inherit: bool = False,
    optimize: int = -1,
    *,
    _feature_version: int = -1,
) -> CodeType: ...


@overload
def _compile(
    source: str | Buffer | ast.AST | ast.AST | ast.AST,
    filename: str | Buffer | PathLike,
    mode: str,
    *,
    dont_inherit: bool = False,
    optimize: int = -1,
    _feature_version: int = -1,
) -> CodeType: ...


@overload
def _compile(
    source: str | Buffer | ast.AST | ast.AST | ast.AST,
    filename: str | Buffer | PathLike,
    mode: str,
    flags: Literal[1024],
    dont_inherit: bool = False,
    optimize: int = -1,
    *,
    _feature_version: int = -1,
) -> ast.AST: ...


@overload
def _compile(
    source: str | Buffer | ast.AST | ast.AST | ast.AST,
    filename: str | Buffer | PathLike,
    mode: str,
    flags: int,
    dont_inherit: bool = False,
    optimize: int = -1,
    *,
    _feature_version: int = -1,
) -> Any: ...


def _delattr(obj: object, name: str, /) -> None: ...


def _dir(o: object = ..., /) -> list: ...


@overload
def _divmod(x: SupportsDivMod, y: Any, /) -> Any: ...


@overload
def _divmod(x: Any, y: SupportsRDivMod, /) -> Any: ...


def _format(value: object, format_spec: str = "", /) -> str: ...


@overload
def _getattr(o: object, name: str, /) -> Any: ...


@overload
def _getattr(o: object, name: str, default: None, /) -> Any: ...


@overload
def _getattr(o: object, name: str, default: bool, /) -> Any | bool: ...


@overload
def _getattr(o: object, name: str, default: list, /) -> Any | list: ...


@overload
def _getattr(o: object, name: str, default: dict, /) -> Any | dict: ...


@overload
def _getattr(o: object, name: str, default: Any, /) -> Any: ...


def _globals() -> dict: ...


def _hasattr(obj: object, name: str, /) -> bool: ...


def _hash(obj: object, /) -> int: ...


def _hex(number: int | int, /) -> str: ...


def _id(obj: object, /) -> int: ...


def _input(prompt: object = "", /) -> str: ...


@overload
def _iter(object: SupportsIter, /) -> Any: ...


@overload
def _iter(object: _GetItemIterable, /) -> Iterator: ...


@overload
def _iter(object: Callable[[], Any], sentinel: None, /) -> Iterator: ...


@overload
def _iter(object: Callable[[], Any], sentinel: object, /) -> Iterator: ...


def _isinstance(obj: object, class_or_tuple: type | tuple, /) -> bool: ...


def _issubclass(cls: type, class_or_tuple: type | tuple, /) -> bool: ...


def _len(obj: Sized, /) -> int: ...


def _locals() -> dict: ...


@overload
def _max(arg1: Any, arg2: Any, /, *_args: Any, key: None = None) -> Any: ...


@overload
def _max(
    arg1: Any, arg2: Any, /, *_args: Any, key: Callable[[Any], SupportsDunderLT | SupportsDunderGT]
) -> Any: ...


@overload
def _max(iterable: Iterable, /, *, key: None = None) -> Any: ...


@overload
def _max(
    iterable: Iterable, /, *, key: Callable[[Any], SupportsDunderLT | SupportsDunderGT]
) -> Any: ...


@overload
def _max(iterable: Iterable, /, *, key: None = None, default: Any) -> Any: ...


@overload
def _max(
    iterable: Iterable,
    /,
    *,
    key: Callable[[Any], SupportsDunderLT | SupportsDunderGT],
    default: Any,
) -> Any: ...


@overload
def _min(arg1: Any, arg2: Any, /, *_args: Any, key: None = None) -> Any: ...


@overload
def _min(
    arg1: Any, arg2: Any, /, *_args: Any, key: Callable[[Any], SupportsDunderLT | SupportsDunderGT]
) -> Any: ...


@overload
def _min(iterable: Iterable, /, *, key: None = None) -> Any: ...


@overload
def _min(
    iterable: Iterable, /, *, key: Callable[[Any], SupportsDunderLT | SupportsDunderGT]
) -> Any: ...


@overload
def _min(iterable: Iterable, /, *, key: None = None, default: Any) -> Any: ...


@overload
def _min(
    iterable: Iterable,
    /,
    *,
    key: Callable[[Any], SupportsDunderLT | SupportsDunderGT],
    default: Any,
) -> Any: ...


@overload
def _next(i: SupportsNext, /) -> Any: ...


@overload
def _next(i: SupportsNext, default: Any, /) -> Any: ...


def _oct(number: int | int, /) -> str: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal[
        "r+",
        "+r",
        "rt+",
        "r+t",
        "+rt",
        "tr+",
        "t+r",
        "+tr",
        "w+",
        "+w",
        "wt+",
        "w+t",
        "+wt",
        "tw+",
        "t+w",
        "+tw",
        "a+",
        "+a",
        "at+",
        "a+t",
        "+at",
        "ta+",
        "t+a",
        "+ta",
        "x+",
        "+x",
        "xt+",
        "x+t",
        "+xt",
        "tx+",
        "t+x",
        "+tx",
    ]
    | Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"]
    | Literal["r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"] = "r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> TextIOWrapper: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal[
        "rb+",
        "r+b",
        "+rb",
        "br+",
        "b+r",
        "+br",
        "wb+",
        "w+b",
        "+wb",
        "bw+",
        "b+w",
        "+bw",
        "ab+",
        "a+b",
        "+ab",
        "ba+",
        "b+a",
        "+ba",
        "xb+",
        "x+b",
        "+xb",
        "bx+",
        "b+x",
        "+bx",
    ]
    | Literal["rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"]
    | Literal["wb", "bw", "ab", "ba", "xb", "bx"],
    buffering: Literal[0],
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> FileIO: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal[
        "rb+",
        "r+b",
        "+rb",
        "br+",
        "b+r",
        "+br",
        "wb+",
        "w+b",
        "+wb",
        "bw+",
        "b+w",
        "+bw",
        "ab+",
        "a+b",
        "+ab",
        "ba+",
        "b+a",
        "+ba",
        "xb+",
        "x+b",
        "+xb",
        "bx+",
        "b+x",
        "+bx",
    ],
    buffering: Literal[-1, 1] = -1,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> BufferedRandom: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal["wb", "bw", "ab", "ba", "xb", "bx"],
    buffering: Literal[-1, 1] = -1,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> BufferedWriter: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal["rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"],
    buffering: Literal[-1, 1] = -1,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> BufferedReader: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: Literal[
        "rb+",
        "r+b",
        "+rb",
        "br+",
        "b+r",
        "+br",
        "wb+",
        "w+b",
        "+wb",
        "bw+",
        "b+w",
        "+bw",
        "ab+",
        "a+b",
        "+ab",
        "ba+",
        "b+a",
        "+ba",
        "xb+",
        "x+b",
        "+xb",
        "bx+",
        "b+x",
        "+bx",
    ]
    | Literal["rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"]
    | Literal["wb", "bw", "ab", "ba", "xb", "bx"],
    buffering: int = -1,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> BinaryIO: ...


@overload
def _open(
    file: int | str | bytes | PathLike[str] | PathLike[bytes],
    mode: str,
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
    closefd: bool = True,
    opener: Callable[[str, int], int] | None = None,
) -> IO[Any]: ...


def _ord(c: str | bytes | bytearray, /) -> int: ...


@overload
def _print(
    *values: object,
    sep: str | None = " ",
    end: str | None = "\n",
    file: SupportsWrite | None = None,
    flush: Literal[False] = False,
) -> None: ...


@overload
def _print(
    *values: object,
    sep: str | None = " ",
    end: str | None = "\n",
    file: _SupportsWriteAndFlush | None = None,
    flush: bool,
) -> None: ...


@overload
def _pow(base: int, exp: int, mod: int) -> int: ...


@overload
def _pow(base: int, exp: Literal[0], mod: None = None) -> Literal[1]: ...


@overload
def _pow(
    base: int,
    exp: Literal[
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25
    ],
    mod: None = None,
) -> int: ...


@overload
def _pow(
    base: int,
    exp: Literal[
        -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -15, -16, -17, -18, -19, -20
    ],
    mod: None = None,
) -> float: ...


@overload
def _pow(base: int, exp: int, mod: None = None) -> Any: ...


@overload
def _pow(
    base: Literal[
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25
    ],
    exp: float,
    mod: None = None,
) -> float: ...


@overload
def _pow(
    base: Literal[
        -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -15, -16, -17, -18, -19, -20
    ],
    exp: float,
    mod: None = None,
) -> complex: ...


@overload
def _pow(base: float, exp: int, mod: None = None) -> float: ...


@overload
def _pow(base: float, exp: complex | _SupportsPow2, mod: None = None) -> Any: ...


@overload
def _pow(base: complex, exp: complex | _SupportsPow2, mod: None = None) -> complex: ...


@overload
def _pow(base: _SupportsPow2, exp: Any, mod: None = None) -> Any: ...


@overload
def _pow(base: _SupportsPow3NoneOnly, exp: Any, mod: None = None) -> Any: ...


@overload
def _pow(base: _SupportsPow3, exp: Any, mod: Any) -> Any: ...


@overload
def _pow(base: _SupportsPow2, exp: float, mod: None = None) -> Any: ...


@overload
def _pow(base: _SupportsPow2, exp: complex, mod: None = None) -> complex: ...


def _repr(obj: object, /) -> str: ...


@overload
def _round(number: _SupportsRound1, ndigits: None = None) -> Any: ...


@overload
def _round(number: _SupportsRound2, ndigits: int) -> Any: ...


def _setattr(obj: object, name: str, value: Any, /) -> None: ...


@overload
def _sorted(iterable: Iterable, /, *, key: None = None, reverse: bool = False) -> list: ...


@overload
def _sorted(
    iterable: Iterable,
    /,
    *,
    key: Callable[[Any], SupportsDunderLT | SupportsDunderGT],
    reverse: bool = False,
) -> list: ...


@overload
def _sum(iterable: Iterable[bool | int], /, start: int = 0) -> int: ...


@overload
def _sum(iterable: Iterable, /) -> Any | Literal[0]: ...


@overload
def _sum(iterable: Iterable, /, start: Any) -> Any: ...


@overload
def _vars(object: type, /) -> types.MappingProxyType[str, Any]: ...


@overload
def _vars(object: Any = ..., /) -> dict: ...



# fmt: off

from .__ import *
from .base import Msg


__all__ = [
    'TermMsg',
    'TermPassByValMsg',
    'TermPassByRefMsg',
    'CONSTANT_ENCODERS',
    'SIMPLE_ENCODERS',
    'COMPLEX_ENCODERS'
]


################################################################################

if TYPE_CHECKING:
    pass

################################################################################

class TermMsg(Msg):
    """ABC for messages defining or referencing terms.

    These can be by-value (ByValueMsg) or by-reference (ResourceMsg).

    TermMsg classes must have the instance method:
    * dec: decode back to a value, requires a MsgDecoder to decode recursively if necessary.

    TermMsg classes must have static methods which will be looked for during class initialization:
    * encode_atomic:     for non-recursive encoding to a ValueMsg
    * encode_compound:   for recursive encoding to a ValueMsg
    * encode_definition: for encoding to a ResourceMsg
    """

    type V = TermT

    def decode(self, dec: DecoderP) -> V:
        raise NotImplementedError(self.__class__.__name__)


################################################################################

class TermPassByValMsg(TermMsg):
    """ABC for messages describing things that serialize by value."""


class TermPassByRefMsg(TermMsg):
    """ABC for messages describing things that serialize by reference."""


################################################################################

CONSTANT_ENCODERS: dict[type, TermMsg] = {}
SIMPLE_ENCODERS:   dict[type, Fn[TermT, TermMsg]] = {}
COMPLEX_ENCODERS:  dict[type, Callable[[TermT, EncoderP], TermMsg]] = {}



# fmt: off

from .log_streams import *
from .log_flag import *
from .log_context import *
from .log_tags import *
from .log_env import *
from .log_base import *

__all__ = [
    'LogFlag',
    'LogContext',
    'LogBase',
    'LoggingSpec',
    'should_log_cls',
    'should_log_tag',
    'cls_log_tags',
    'set_log_tags',
    'set_log_streams',
    'binary_log_dir',
    'ScopedLogging',
    'IN_WASM',
]



# fmt: off

from builtins import type as _type
from typing import Any

from ...cpython.classes.anno import TGeneric
from .. import flags
from ..attrs import VHDL
from ..request.request_resource import *
from .__raw import *
from .base import *

__all__ = [
    'V_OBJECT_METHODS',
    'V_CLASS_GETITEM',
]


################################################################################

PY_OBJECT: type[object] = object

# TODO:
# - modifying the dic returned by __dict__ should trigger __setattr__ on the remote;
# - trying to set `__dict__` should trigger the remote to replace its attributes.

def __dict__(self) -> dict[str, Any]:
    handle = obj_handle(self)
    if handle.open:
        return handle.hdlr(handle, ResourceCallSystemMethod(self, vars))
    else:
        dunder_dict = {}
        for k in handle.keys:
            try:
                dunder_dict[k] = getattr(self, k)
            except AttributeError:
                pass
        return dunder_dict


class object:

    def __str__(self) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallSystemMethod(self, str))

    def __repr__(self) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallSystemMethod(self, repr))

    def __hash__(self) -> Any:
        handle = obj_handle(self)
        if hasattr(handle, 'hash'):
            return handle.hash
        result = handle.hdlr(handle, ResourceCallSystemMethod(self, hash))
        if _type(result) is int:
            setattr(handle, 'hash', result)
            return result
        return id(self)

    def __getattribute__(self, name: str, /) -> Any:
        if flags.VIRTUAL_OBJECT_DUNDER_DICT and name == '__dict__':
            return __dict__(self)
        return PY_OBJECT.__getattribute__(self, name)

    def __getattr__(self, name: str) -> Any:
        handle = obj_handle(self)
        if not handle.open and name not in handle.keys and not hasattr(_type(self), name):
            raise AttributeError(f"{handle.name} has no attribute {name!r}")
        value = handle.hdlr(handle, ResourceGetAttr(self, name))
        if name not in handle.keys:
            handle.keys.append(name)
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        handle = obj_handle(self)
        handle.hdlr(handle, ResourceSetAttr(self, name, value))
        keys = handle.keys
        if name not in keys:
            keys.append(name)

    def __delattr__(self, name: str):
        handle = obj_handle(self)
        if handle.open:
            handle.hdlr(handle, ResourceDelAttr(self, name))
            if name in handle.keys:
                handle.keys.remove(name)
        else:
            if name not in handle.keys:
                raise AttributeError(f"{handle.name} has no attribute {name!r}")
            handle.hdlr(handle, ResourceDelAttr(self, name))
            handle.keys.remove(name)

    def __dir__(self) -> list[str]:
        handle = obj_handle(self)
        if handle.open:
            return handle.hdlr(handle, ResourceCallSystemMethod(self, dir))
        else:
            c_dir = dir(_type(self))
            if VHDL in c_dir:
                c_dir.remove(VHDL)
            return handle.keys + c_dir


V_OBJECT_METHODS = {
    '__str__':     object.__str__,
    '__repr__':    object.__repr__,
    '__dir__':     object.__dir__,
    '__hash__':    object.__hash__,
    '__getattr__': object.__getattr__,
    '__setattr__': object.__setattr__,
    '__delattr__': object.__delattr__,
    # Put __format__ on here so it doesn't cause RPC
    '__format__':  PY_OBJECT.__format__,
}

if flags.VIRTUAL_OBJECT_DUNDER_DICT:
    V_OBJECT_METHODS['__getattribute__'] = object.__getattribute__
    V_OBJECT_METHODS['__dict__'] = property(__dict__)

if not flags.VIRTUAL_OBJECT_MUTATION:
    del V_OBJECT_METHODS['__setattr__']
    del V_OBJECT_METHODS['__delattr__']

################################################################################


class type:
    def __class_getitem__(cls, item):
        return TGeneric(cls, item)


V_CLASS_GETITEM = type.__dict__['__class_getitem__']



# fmt: off

from typing import Any, Literal
from dataclasses import dataclass
from types import NoneType, NotImplementedType, EllipsisType, ModuleType

from ..core.type import PRIMITIVES, CALLABLES, CONTAINERS
from ..core.anno import ANNOS, anno_str, is_anno_class
from ..core.fmt import f_callable

__all__ = [
    'ReplVarInfo',
    'VarKind',
    'var_kind'
]


################################################################################

@dataclass
class ReplVarInfo:
    """
    Encapsulates *serializable* information about a variable + value in the REPL.
    """

    cls_name:     str
    is_class:     bool
    is_type_anno: bool
    is_callable:  bool
    safe_repr:    str | None
    kind:        'VarKind'

    @staticmethod
    def from_value(value: Any, /) -> 'ReplVarInfo':

        value_cls = type(value)
        is_class = value_cls is type or isinstance(value, type)
        is_type_anno = is_anno_class(value_cls)

        cls_module = getattr(value_cls, '__module__', None)
        cls_name = value_cls.__name__
        if type(cls_module) is str:
            cls_name = f'{cls_module}.{cls_name}'

        is_callable = callable(value)

        if is_class or is_type_anno:
            safe_repr = anno_str(value)
        elif value_cls in REPR_SAFE_TYPES:
            safe_repr = repr(value)
        elif value_cls in CALLABLES:
            safe_repr = f_callable(value)
        else:
            safe_repr = None

        kind = var_kind(value)

        return ReplVarInfo(
            cls_name=cls_name,
            is_type_anno=is_type_anno,
            is_class=is_class,
            is_callable=is_callable,
            safe_repr=safe_repr,
            kind=kind,
        )

    @property
    def is_type_like(self) -> bool:
        return self.is_class or self.is_type_anno


REPR_SAFE_TYPES = bool, int, float, NoneType, NotImplementedType, EllipsisType, ModuleType, type

################################################################################

type VarKind = Literal['data', 'object', 'class', 'type', 'function', 'module', 'future']

DATA = PRIMITIVES + CONTAINERS + (str, bytes)

def var_kind(obj: Any) -> VarKind:
    cls = type(obj)
    if issubclass(cls, DATA):
        return 'data'
    elif cls is type or isinstance(obj, type):
        return 'class'
    elif cls in ANNOS:
        return 'type'
    elif cls in CALLABLES or callable(obj):
        return 'function'
    elif issubclass(cls, ModuleType):
        return 'module'
    elif cls.__module__ == 'asyncio' and cls.__name__ == 'Future':
        return 'future'
    else:
        return 'object'



# fmt: off

from typing import ClassVar

__all__ = [
    'CvarClassBase',
    'CvarClass',
    'CvarClassSub',
    'WITH_CVAR_CLASSES',
]


class CvarClassBase:
    """CvarClassBase docstring"""

    a: ClassVar[int] = 0


class CvarClass(CvarClassBase):
    """CvarClass docstring"""

    b: ClassVar[bool]
    c: ClassVar[str] = "foo"


class CvarClassSub(CvarClass):
    """CvarClassSub docstring"""

    a: ClassVar[int] = 5  # type: ignore
    d: ClassVar[bytes | None]


WITH_CVAR_CLASSES: list[type] = [
    CvarClassBase,
    CvarClass,
    CvarClassSub,
]



# fmt: off

from .builtin import BUILTIN_CLASSES
from .dclasses import DCLASS_CLASSES
from .diamonds import DIAMOND_CLASSES
from .enums import ENUM_CLASSES
from .generic import GENERIC_CLASSES
from .named_tuples import NAMED_TUPLE_CLASSES
from .recursive import RECURSIVE_CLASSES
from .typed_dicts import TYPED_DICT_CLASSES
from .units import UNIT_CLASSES
from .with_annos import WITH_ANNO_CLASSES
from .with_cvars import WITH_CVAR_CLASSES
from .with_methods import WITH_METHOD_CLASSES
from .with_props import WITH_PROP_CLASSES

__all__ = [
    'BUILTIN_CLASSES',
    'DCLASS_CLASSES',
    'DIAMOND_CLASSES',
    'ENUM_CLASSES',
    'NAMED_TUPLE_CLASSES',
    'RECURSIVE_CLASSES',
    'TYPED_DICT_CLASSES',
    'GENERIC_CLASSES',
    'WITH_ANNO_CLASSES',
    'WITH_CVAR_CLASSES',
    'WITH_METHOD_CLASSES',
    'WITH_PROP_CLASSES',
    'CLASSES',
]

CLASSES = [
    *DCLASS_CLASSES,
    *DIAMOND_CLASSES,
    *ENUM_CLASSES,
    *NAMED_TUPLE_CLASSES,
    *RECURSIVE_CLASSES,
    *TYPED_DICT_CLASSES,
    *GENERIC_CLASSES,
    *WITH_ANNO_CLASSES,
    *WITH_METHOD_CLASSES,
    *WITH_PROP_CLASSES,
    *UNIT_CLASSES
]



# fmt: off

from .__ import *
from .base import *


__all__ = [
    'ModuleData',
]


################################################################################

class ModuleData(ResourceData):
    __slots__ = 'name', 'doc', 'file', 'exports', 'keys', 'annos'

    FORBIDDEN_FORM = forbidden_module

    name:    str
    doc:     optstr
    file:    optstr
    exports: ResourcesRecordT
    keys:    strtup
    annos:   AnnotationsT

    def create_resource(self, handle: ResourceHandle) -> ModuleType:
        handle.kind = Kind.Module
        exports = self.exports

        handle = new_handle(context, 'module', self.keys)

        meta = {
            FILE: self.file,
            ALL: list(exports.keys()),
            ANNOS: self.annos,
            VHDL: handle
        }

        return module(self.name, self.doc, **meta)


################################################################################

def describe_real_module(mod: ModuleT) -> ModuleData:

    name, file, doc, annos, syms = multi_get_raw(mod, NAME, FILE, DOC, ANNOS, ALL)

    assert is_str(name), "name must be a string"
    assert is_optstr(file), "file must be a string or None"
    assert is_optstr(doc), "doc must be a string or None"
    assert is_rec(annos), "annos must be a record"
    assert syms is None or is_strlist(syms), "exports must be a list of strings or None"

    if is_forbidden(mod, name):
        raise E.WarpEncodingForbiddenError(f"<module {name!r}>")

    data = ModuleData()

    data.name = name
    data.file = file
    data.doc = doc
    data.annos = annos

    data.keys = keys = tuple(mod.__dir__())
    data.exports, add_export = mkdict()
    for key in syms or ():
        if key not in keys:
            continue
        val = getattr(mod, key, FIELD_ABSENT)
        if val is not FIELD_ABSENT:
            add_export(key, val)

    return data


################################################################################

class module(ModuleType):

    def __init__(self, name: str, doc: str | None, **kwargs) -> None:
        super().__init__(name, doc)
        for k, v in kwargs.items():
            mod_set(self, k, v)

    def __getattr__(self, name: str) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceGetAttr(self, name))

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError('this module is immutable')

    def __delattr__(self, name: str):
        raise TypeError('this module is immutable')

    def __dir__(self) -> list[str]:
        return obj_handle(self).keys

module.__module__ = 'virtual'

################################################################################

def create_virtual_module(data: ModuleData, handle: ResourceHandle) -> ModuleType:

    handle.kind = Kind.Module
    handle.keys = list(data.keys)
    handle.open = False
    handle.name = data.name

    exports = data.exports

    meta = {
        FILE: data.file,
        ALL: list(exports.keys()),
        ANNOS: data.annos,
        VHDL: handle
    }

    return module(data.name, data.doc, **meta)


################################################################################

ModuleData.describe_resource = staticmethod(describe_real_module)
ModuleData.create_resource = create_virtual_module



import asyncio

from agentica_internal.warpc.worlds.all import DebugWorld


class MyFunkyException(BaseException):
    pass


def raise_remote():
    raise MyFunkyException('hello!')


async def run():
    async with DebugWorld.connected_pair(logging='ENCR+DECR') as B:
        fn = await B(raise_remote)
        await fn()


asyncio.run(run(), debug=True)



# fmt: off

from os import getenv, listdir, environ
from sys import __stderr__ as STDERR
from collections.abc import Callable
from contextvars import ContextVar

from .log_streams import set_log_streams

__all__ = [
    'get_log_tags',
    'set_log_tags',
    'ScopedLogging',
    'LoggingSpec',
    'IN_WASM'
]


###############################################################################

FROM_ENV = object()

CUR_LOG_TAGS = ContextVar('AGENTICA_LOG_TAGS', default=FROM_ENV)

def get_log_tags() -> str | None:
    tags = CUR_LOG_TAGS.get()
    if tags is FROM_ENV:
        tags = getenv('AGENTICA_LOG_TAGS')
    return tags

###############################################################################

type LoggingSpec = bool | str | set[str] | None

def set_log_tags(spec: LoggingSpec, is_global: bool = False) -> Callable[[], None]:
    env = None
    tags = parse_log_tags_spec(spec)
    if is_global:
        env = environ.get('AGENTICA_LOG_TAGS')
        environ['AGENTICA_LOG_TAGS'] = tags
    if tags is None:
        return _noop
    if tags.startswith('+'):
        curr = get_log_tags()
        tags = _add_tags(curr, tags)
    token = CUR_LOG_TAGS.set(tags)
    _refresh()
    def reset() -> None:
        if is_global:
            if env is None:
                del environ['AGENTICA_LOG_TAGS']
            else:
                environ['AGENTICA_LOG_TAGS'] = env
        CUR_LOG_TAGS.reset(token)
        _refresh()
    return reset

def _noop() -> None:
    pass

def _refresh() -> None:
    from .log_base import refresh_log_bases
    from .log_flag import FLAGS
    for flag in FLAGS:
        flag.flag = None
    refresh_log_bases()

def _add_tags(old: str | None, new: str) -> str:
    if not old or old == '0':
        return new
    if not new:
        return old
    if old == '1' or old == 'ALL':
        return old
    return old + '+' + new

################################################################################

class ScopedLogging:
    tag_spec: LoggingSpec
    log_files: list[str]
    reset_env: Callable[[], None] | None
    reset_tags: Callable[[], None] | None
    reset_io: Callable[[], None] | None
    is_global: bool
    stdout: bool
    stderr: bool

    def __init__(self, tag_spec: LoggingSpec, *log_files: str, stdout: bool = True, stderr: bool = True,
                 is_global: bool = False, ):
        self.tag_spec = tag_spec
        self.log_files = list(log_files)
        self.reset_tags = None
        self.reset_io = None
        self.is_global = is_global
        self.stdout = stdout
        self.stderr = stderr

    def __enter__(self):
        self.begin()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    def begin(self):
        STDERR.write(f"\nTURNING ON {self.tag_spec} LOGGING\n")
        self.reset_tags = set_log_tags(self.tag_spec, self.is_global)
        self.reset_io = set_log_streams(*self.log_files, stdout=self.stdout, stderr=self.stderr)
        STDERR.flush()

    def end(self):
        if reset_tags := self.reset_tags:
            STDERR.write(f"\nTURNING OFF {self.tag_spec} LOGGING\n")
            reset_tags()
        if reset_io := self.reset_io:
            STDERR.write("\nRESETTING IO\n")
            reset_io()
        STDERR.flush()


def parse_log_tags_spec(spec: LoggingSpec) -> str | None:
    match spec:
        case None:
            return None
        case True:
            return '1'
        case False:
            return '0'
        case 'ALL':
            return 'ALL'
        case str():
            return spec
        case set():
            return '+'.join(spec)
        case _:
            STDERR.write(f"UNKNOWN FORCE LOGGING SPEC: {spec!r}\n")
            return None

###############################################################################

try:
    listdir()
    IN_WASM = False
except:
    IN_WASM = True



import typing as T
from importlib import import_module
from linecache import getline
from pathlib import Path
from sys import modules as sys_modules
from types import FunctionType, ModuleType

__all__ = [
    'get_shed_module',
    'get_shed_function',
    'get_shed_class',
    'get_shed_method',
]


THIS_DIR = Path(__file__).parent


NOT_THERE = object()

_original_overload = T.overload


# this is like overload, but doesn't return _overload_dummy, since this ruins typeshed import
def _patched_overload(func: T.Callable) -> T.Callable:
    f = getattr(func, "__func__", func)
    try:
        T._overload_registry[f.__module__][f.__qualname__][f.__code__.co_firstlineno] = func  # type: ignore
    except AttributeError:
        # Not a normal function; ignore.
        pass
    return func


def process_module(module: ModuleType, prefix: str = '') -> ModuleType:
    assert hasattr(module, '__all__')

    _names = list(module.__all__)
    if prefix:
        _names = [n for n in _names if n.startswith(prefix)]

    _modname = module.__name__
    modname = _modname.removeprefix('agentica_internal.cpython.shed._')
    file = module.__file__

    functions = []
    add_function = functions.append

    classes = []
    add_class = classes.append

    names = []
    addname = names.append
    newdict = {}
    additem = newdict.__setitem__

    sys_mod: ModuleType = sys_modules.get(modname, NotImplemented)

    getname = module.__dict__.get
    for _name in _names:
        name = _name
        if prefix:
            if not _name.startswith(prefix):
                continue
            name = _name.removeprefix(prefix)

        obj = getname(_name, NOT_THERE)
        if obj is NOT_THERE:
            continue

        sys_obj = getattr(sys_mod, name, NotImplemented)

        if isinstance(obj, FunctionType):
            obj.___shed_obj___ = True
            obj.___shed_ori___ = sys_obj
            obj.___shed_src___ = fun_source(obj, file)
            obj.__module__ = modname
            obj.__qualname__ = obj.__name__ = name
            add_function(obj)

        elif isinstance(obj, type):
            obj.___shed_obj___ = True
            obj.___shed_ori___ = sys_obj
            obj.___shed_src___ = cls_source(obj, file)
            obj.__module__ = modname
            obj.__qualname__ = obj.__name__ = name
            for fieldname, field in obj.__dict__.items():
                if isinstance(field, (classmethod, staticmethod)):
                    field = field.__wrapped__
                    sys_obj = getattr(sys_obj, '__wrapped__', NotImplemented)
                if isinstance(field, FunctionType):
                    field.___shed_obj___ = True
                    field.___shed_ori___ = sys_obj
                    field.___shed_src___ = fun_source(field, file)
                    field.__module__ = modname
                    field.__qualname__ = f'{name}.{fieldname}'
            add_class(obj)
        additem(name, obj)
        addname(name)

    newdict['__all__'] = names
    newdict['__CLASSES__'] = classes
    newdict['__FUNCTIONS__'] = functions
    newdict['___shed_obj___'] = True
    newdict['___shed_ori___'] = sys_mod
    newmod = ModuleType(module.__name__)
    newmod.__dict__.update(newdict)
    return newmod


def cls_source(cls: type, file: str):
    lines = []
    lineno = 1
    header = f'class {cls.__name__}'
    while not getline(file, lineno).startswith(header):
        lineno += 1
    while True:
        line = getline(file, lineno)
        if line.startswith('##'):
            break
        lines.append(line)
        lineno += 1
    return '\n'.join(lines).strip()


def fun_source(fun: FunctionType, file: str):
    lineiter = fun.__code__.co_lines()
    linenos = [n for s, e, n in lineiter if type(n) is int]
    start = min(linenos)
    end = max(linenos)
    lines = [getline(file, lineno) for lineno in range(start, end + 1)]
    return '\n'.join(lines)


class _ShedModuleCache(dict[str, ModuleType]):
    def __missing__(self, mod_name: str):
        key = mod_name
        mod_name = mod_name.replace('.', '_')
        mod_name = f'agentica_internal.cpython.shed._{mod_name}'
        if mod_name in sys_modules:
            del sys_modules[mod_name]
        try:
            prefix = '_' if key == 'builtins' else ''
            T.overload = _patched_overload  # type: ignore
            mod = import_module(mod_name)
            mod = process_module(mod, prefix)
            self[key] = mod
            return mod
        except ImportError:
            T.overload = _original_overload
            pass
        return FALLBACK_MODULE


class _ShedFunctionCache(dict[tuple[str, str], FunctionType]):
    def __missing__(self, tup: tuple[str, str]):
        mod_name, func_name = tup
        mod = _shed_module_cache[mod_name]
        if mod is not FALLBACK_MODULE:
            func = mod.__dict__.get(func_name, FALLBACK_FUNCTION)
            if func is not FALLBACK_FUNCTION:
                return func

        def fallback_func(*args, **kwargs): ...

        fallback_func.__module__ = mod_name
        fallback_func.__name__ = fallback_func.__qualname__ = func_name
        return fallback_func


class _ShedClassCache(dict[type, type]):
    def __missing__(self, cls: type):
        mod = _shed_module_cache[cls.__module__]
        if mod is FALLBACK_MODULE:
            return FALLBACK_CLASS
        attr = mod.__dict__.get(cls.__name__, FALLBACK_CLASS)
        if attr is FALLBACK_CLASS:
            return FALLBACK_CLASS
        attr.__doc__ = cls.__doc__
        assert isinstance(attr, type)
        assert not (attr.__flags__ & 256)
        return attr


RawMethod = FunctionType | staticmethod | classmethod


class _ShedMethodCache(dict[tuple[type, str], RawMethod]):
    def __missing__(self, tup: tuple[type, str]):
        cls, name = tup
        shed_cls = _shed_class_cache[cls]
        if shed_cls is not FALLBACK_CLASS:
            meth = shed_cls.__dict__.get(name, FALLBACK_METHOD)
            if meth is not FALLBACK_METHOD and isinstance(meth, RAW_METHOD_TYPES):
                return meth

        def fallback_meth(*args, **kwargs): ...

        fallback_meth.__module__ = cls.__module__
        fallback_meth.__name__ = name
        fallback_meth.__qualname__ = f'{cls.__name__}.{name}'
        return fallback_meth


RAW_METHOD_TYPES = (FunctionType, staticmethod, classmethod)


_shed_module_cache = _ShedModuleCache()
_shed_function_cache = _ShedFunctionCache()
_shed_class_cache = _ShedClassCache()
_shed_method_cache = _ShedMethodCache()


def get_shed_module(mod_name: str):
    return _shed_module_cache[mod_name]


def get_shed_function(mod_name: str, func_name: str) -> FunctionType:
    key = mod_name, func_name
    return _shed_function_cache[key]


def get_shed_class(cls: type) -> type:
    return _shed_class_cache[cls]


# this is the *raw* function, straight from the dict
def get_shed_method(cls: type, meth_name: str) -> RawMethod:
    key = cls, meth_name
    return _shed_method_cache[key]


FALLBACK_MODULE = ModuleType('<notfound>')


class FALLBACK_CLASS:
    def FALLBACK_METHOD(self, *args, **kwargs):
        pass


FALLBACK_METHOD = FALLBACK_CLASS.__dict__['FALLBACK_METHOD']


def FALLBACK_FUNCTION(*args, **kwargs):
    pass



from dis import get_instructions
from operator import attrgetter
from types import CodeType
from typing import NamedTuple, TypeGuard

__all__ = [
    'is_code',
    'is_async_code',
    'code_arg_info',
    'code_arg_names',
    'print_code',
    'CodeArgInfo',
    'FLAGS',
]


def is_code(obj: object) -> TypeGuard[CodeType]:
    """Returns True for `CodeType` objects."""

    return isinstance(obj, CodeType)


def is_async_code(code: CodeType) -> bool:
    return isinstance(code, CodeType) and bool(code.co_flags & FLAGS.IS_ASYNC)


class CodeArgInfo(NamedTuple):
    """A 6-tuple describing a CodeType object:

    0. `var_names` | tuple of argument names followed by local var names
    1. `num_pos`   | number of positional-only arguments
    2. `num_reg`   | number of positional-or-keyword arguments
    3. `num_key`   | number of keyword-only arguments
    4. `pos_star`  | True if there is a final *args argument
    5. `key_star`  | True if there is a final **kwargs argument

    Note: var_names includes locals, arg_count excludes * and ** args
    """

    var_names: tuple[str, ...]
    num_pos: int
    num_reg: int
    num_key: int
    pos_star: bool
    key_star: bool


codeattrs = attrgetter(
    'co_varnames', 'co_flags', 'co_argcount', 'co_posonlyargcount', 'co_kwonlyargcount'
)


def code_arg_info(code: CodeType) -> CodeArgInfo:
    """Given a CodeType object, returns a CodeArgInfo describing it properties."""

    assert type(code) is CodeType, f'{code} is not a CodeType object'

    varnames, flags, arg_count, num_pos, num_key = codeattrs(code)

    return CodeArgInfo(
        varnames,
        num_pos,
        arg_count - num_pos,
        num_key,
        bool(flags & FLAGS.VARARGS),
        bool(flags & FLAGS.VARKEYWORDS),
    )


def code_arg_names(code: CodeType) -> list[str]:
    """Given a CodeType object, returns its argument names."""

    assert type(code) is CodeType, f'{code} is not a CodeType object'

    varnames, flags, arg_count, num_pos, num_key = codeattrs(code)
    pos_star = bool(flags & FLAGS.VARARGS)
    key_star = bool(flags & FLAGS.VARARGS)

    return varnames[: arg_count + num_key + pos_star + key_star]


class FLAGS:
    """Flags present on a CodeObject."""

    OPTIMIZED = 1
    NEWLOCALS = 2
    VARARGS = 4
    VARKEYWORDS = 8
    NESTED = 16
    GENERATOR = 32
    NOFREE = 64
    COROUTINE = 128
    ITERABLE_COROUTINE = 256
    ASYNC_GENERATOR = 512
    IS_ASYNC = ASYNC_GENERATOR | COROUTINE | ITERABLE_COROUTINE


def print_code(code: CodeType) -> None:
    print(f'''
CodeType(
    co_name     = {code.co_name!r}
    co_qualname = {code.co_qualname!r}
    co_argcount = {code.co_argcount}
    co_nlocals  = {code.co_nlocals}
    co_names    = {fmt_names(code.co_names)}
    co_varnames = {fmt_names(code.co_varnames)}
    co_freevars = {fmt_names(code.co_freevars)}
    co_cellvars = {fmt_names(code.co_cellvars)}
    co_conts    = {code.co_consts}
    co_flags    = {code.co_flags}
    co_code     = {code.co_code!r}
                  {fmt_bytecode(code)}>
)''')


def fmt_names(strs: tuple[str, ...]) -> str:
    return '[' + ', '.join(map(repr, strs)) + ']'


def fmt_bytecode(code: CodeType) -> str:
    strs = []
    for i, ins in enumerate(get_instructions(code)):
        strs.append(f'{ins.opname}({ins.arg})')
        if i > 8:
            break
    return '[' + ', '.join(strs) + ']'



# fmt: off

from collections.abc import Callable
from .log_tags import *
from .log_fns import *

__all__ = [
    'LogFlag',
    'FLAGS'
]

###############################################################################

class LogFlag:
    tags: set[str]
    flag: bool | None

    def __init__(self, spec: type | str):
        if type(spec) is str:
            self.tags = set(spec.lower().split('+'))
        elif isinstance(spec, type):
            self.tags = cls_log_tags(spec)
        else:
            raise TypeError(spec)
        self.flag = None
        FLAGS.append(self)

    def __bool__(self):
        flag = self.flag
        if flag is None:
            flag = should_log_tag(False, self.tags)
            self.flag = flag
        return flag

    def __repr__(self):
        f_tags = '+'.join(self.tags)
        return f'<logflag {f_tags} {self.flag}>'

    def log_fn(self, *args) -> Callable[..., None]:
        return log_fn(args) if self else no_log_fn

    def enable(self) -> None:
        self.flag = True

    def disable(self) -> None:
        self.flag = False


FLAGS: list[LogFlag] = []



# fmt: off

import sys

from sys import modules as sys_modules
from asyncio import Future, Task, AbstractEventLoop, sleep, _get_running_loop
from collections import deque

from .__ import *
from .base_world import *

__all__ = ['AgentWorld']

################################################################################

PROTECTED_MODULES = set(sys.modules.keys())

class AgentWorld(World):

    recv_bytes: SyncRecvBytes
    send_bytes: SyncSendBytes
    responses:  dict[MessageID, FramedResponseMsg] = {}
    awaiting:   dict[MessageID, Future[FramedResponseMsg]] = {}
    futures:    dict[FutureID, Future] = {}
    channel:    deque[ChannelMsg] | None
    running:    bool
    loop:       AbstractEventLoop
    task:       Task | None

    dummy_modules: dict[str, ModuleType]

    protected_modules: set[str] = set()

    ############################################################################

    def __post_init__(self) -> None:
        self.running = False
        self.channel = deque()
        self.responses = {}
        self.awaiting = {}
        self.futures = {}
        self.task = None
        # modules that exist when the world is created are protected
        self.protected_modules |= PROTECTED_MODULES
        self.dummy_modules = {}

    def _closed(self):
        self.running = False
        self.channel = None
        self.responses.clear()
        self.awaiting.clear()
        self.futures.clear()

    ############################################################################

    def set_loop(self, loop: AbstractEventLoop):
        self.loop = loop
        if loop.is_running():
            raise RuntimeError(f"AgentWorld given a running loop {loop}")
        self.log("loop set to", loop)

    ############################################################################

    def send_msg(self, msg: RPCMsg):
        try:
            data = msg.to_msgpack()
            self.send_bytes(data)
        except Exception as error:
            self.log_forced('send_msg failed; raising WarpShutdown', error)
            raise E.WarpShutdown()

    def __recv_msg(self, ctx: LogContext) -> RPCMsg:
        try:
            ctx.info('awaiting bytes')
            data = self.recv_bytes()
        except RuntimeError:
            ctx.log('recv_bytes broken, exiting')
            raise E.WarpShutdown()
        if data == QUIT:
            ctx.log('QUIT received, exiting')
            raise E.WarpShutdown()
        try:
            rpc_msg = RPCMsg.from_msgpack(data)
            return rpc_msg
        except:
            ctx.warn('corrupt message, exciting')
            raise E.WarpShutdown()

    def __pop_queue(self) -> Result | Closed:
        channel_msg = self.channel.popleft()
        if channel_msg.last:
            self.channel = None
        return channel_msg.decode(self.root)

    ############################################################################

    def resource_post_decode(self, resource: ResourceT, handle: ResourceHandle) -> None:
        # fake the existence of the resource in the correct module

        mod_name = getattr(resource, '__module__', None)
        if type(mod_name) is not str or not mod_name:
            return

        obj_name: str | None = getattr(handle, 'name', None) or getattr(resource, '__name__', None)
        if type(obj_name) is not str or not obj_name:
            return

        dummy_modules = self.dummy_modules
        if mod_name in sys_modules and mod_name not in dummy_modules:
            # if module exists in system modules and is not one we made,
            # then do nothing
            return

        module = dummy_modules.get(mod_name)
        if module is None:
            dummy_modules[mod_name] = module = ModuleType(mod_name, "remote module")

        if mod_name not in sys_modules:
            sys_modules[mod_name] = module

        setattr(module, obj_name, resource)

    ############################################################################

    # to satisfy WorldP

    def execute_outgoing_request_sync(self, msg: FramedRequestMsg) -> FramedResponseMsg:
        with self.log_as('execute_outgoing_request_sync', msg.mid) as ctx:
            ctx.log('sending msg', msg)
            self.send_msg(msg)
            ctx.log('message sent')
            return self.run_until_mid(ctx, msg.mid)

    async def execute_outgoing_request_coro(self, msg: FramedRequestMsg) -> FramedResponseMsg:

        with self.log_as('execute_outgoing_request_coro', msg) as ctx:
            self._check_loop(ctx)

            # otherwise, we start an asyncio task that will service messages
            # and this task will at some point cause the future to be done
            # it is fine for this task to 'compete' with run_until_mid,
            # because read_msg is blocking, only one of them will ever be active
            # on the task queue, and it does the same job as the other

            # if such a task is already running, we do nothing

            ctx.log('sending msg')
            self.send_msg(msg)

            self.start_futures_task()

            mid = msg.mid
            self.awaiting[mid] = future = self.loop.create_future()
            ctx.log(f'future #{f_id(mid)}', future, 'created')

            return await future

    def execute_outgoing_request_future(self, msg: FramedRequestMsg) -> Future[FramedResponseMsg]:

        with self.log_as('execute_outgoing_request_future', msg) as ctx:
            ctx.log('sending msg')
            self.send_msg(msg)

            self.start_futures_task()

            mid = msg.mid
            self.awaiting[mid] = future = self.loop.create_future()
            ctx.log(f'future #{f_id(mid)}', future, 'created')

            return future

    ############################################################################

    def _new_future(self, future_id: FutureID, /) -> FutureT | None:
        future = new_hookable_future(self.loop)
        setattr(future, FUTURE_ID, future_id)
        return future

    def _await_future(self, future: FutureT, _: None):
        with self.log_as('_await_future', future) as ctx:
            self._check_future(future, ctx)
            self.start_futures_task()

    def _gather_future(self, future: FutureT, _: None):
        with self.log_as('_gather_future', future) as ctx:
            self._check_future(future, ctx)
            self.start_futures_task()

    def send_event(self, msg: EventMsg) -> None:
        self.on_event(Outgoing, self.tick, msg)
        self.send_msg(msg)

    def _check_future(self, future: FutureT, ctx: LogContext):
        future_loop = future.get_loop()
        our_loop = self.loop
        if future_loop is not our_loop:
            self._raise_async_error(ctx, 'future', future, 'on loop', future_loop, 'instead of', our_loop)

    def _check_loop(self, ctx: LogContext):
        our_loop = self.loop
        if not our_loop:
            self._raise_async_error(ctx, 'no loop set')
        running_loop = _get_running_loop()
        if running_loop and running_loop is not our_loop:
            self._raise_async_error(ctx, 'running loop', running_loop, 'is not our loop', our_loop)

    def _raise_async_error(self, ctx: LogContext, *args) -> NoReturn:
        ctx.log(*args)
        raise E.WarpAsyncError(P.sprint(*args))

    ############################################################################

    def run_msg_loop(self,
            send_bytes: SyncSendBytes,
            recv_bytes: SyncRecvBytes,
            recv_ready: SyncRecvReady) -> None:

        del recv_ready

        # run a message loop forever, this is the outermost kind of message loop,
        # and we might have many run_until_mid appearing on the stack
        # multiple times AND have `futures_task` running if asyncio gets involved,
        # this is all fine

        with self.log_as('run_msg_loop', self.loop) as ctx:

            self._check_loop(ctx)
            self.recv_bytes = recv_bytes
            self.send_bytes = send_bytes
            try:
                self.run_until_mid(ctx, NO_MID)  # we don't expect any replies at top-level
            except E.WarpShutdown:
                ctx.log('caught WarpShutdown; shutting down')
                pass

    ############################################################################

    def run_until_mid(self, ctx: LogContext, until_mid: MessageID, break_fn=None) -> FramedResponseMsg:

        # this is the 'hand-rolled async' version of the AgentWorld message loop,
        # it runs a given message ID is received, but in the meantime stores any
        # responses that loops higher up in the stack frame may be interested in,
        # and fulfills any active futures (which would imply we are in operating
        # in hybrid mode)

        # compare with `futures_task`

        futures = self.awaiting
        responses = self.responses
        frame = self.root
        loop = self.loop
        send = self.send_msg

        log = bool(ctx)

        while True:

            # if another nested call already received this response
            if response := self.responses.pop(until_mid, None):
                ctx.log('mid=', until_mid, 'received')
                return response

            if break_fn and break_fn():
                ctx.log('break_fn stopped loop')
                return None  # type: ignore

            msg: RPCMsg = self.__recv_msg(ctx)
            ctx.log('got message', msg)

            if isinstance(msg, FramedRequestMsg):
                exit_fn = self.on_frame(Incoming, msg)
                ctx.log('is_async =', msg.is_async)
                if msg.is_async and loop.is_running():
                    ctx.log('calling frame.exec_incoming_request_task')
                    frame.exec_incoming_request_task(msg, loop, send, exit_fn)
                else:
                    ctx.log('calling frame.exec_incoming_request')
                    frame.exec_incoming_request(msg, send, exit_fn)

            elif isinstance(msg, FramedResponseMsg):
                received_mid = msg.mid
                if future := futures.get(received_mid):
                    if not future.done():
                        ctx.log(f'future #{f_id(received_mid)} fulfilled:', future)
                        future.set_result(msg)
                    else:
                        ctx.warn(f'future #{f_id(received_mid)} already set:', future, future._state)
                    del futures[received_mid]

                if received_mid == until_mid:
                    ctx.log('returning matched response, mid =', received_mid)
                    return msg
                elif not future:
                    ctx.log('saving unmatched response, mid =', received_mid)
                    responses[received_mid] = msg

            elif isinstance(msg, FutureEventMsg):
                self.handle_future_event_msg(msg)

            elif isinstance(msg, ChannelMsg):
                channel = self.channel
                if channel is not None:
                    channel.append(msg)

            else:
                self.raise_protocol_error(ctx, f'unexpected message: {msg}')

        raise RuntimeError('unreachable')

    ############################################################################

    def start_futures_task(self) -> None:
        if self.task:
            return
        # self.log('starting futures task')
        name = f'{self.log_name}.futures_task'
        self.task = self.loop.create_task(self.futures_task(), name=name)

    def __log_futures_task_status(self, ctx: LogContext) -> None:
        if awaiting := self.awaiting:
            ctx.log('still awaiting keys:', ' '.join(f_id(k) for k in awaiting.keys()))
        if futures := self.futures:
            ctx.log('still awaiting futures: ', list(futures.keys()))
        if not futures and not awaiting:
            ctx.log('no further reason for futures task')

    async def futures_task(self) -> None:
        # this is the 'asyncio' version of the AgentWorld message loop, it runs
        # as long as there are Futures that can have the tasks waiting on them
        # awoken when we receive their replies

        loop = self.loop
        awaiting = self.awaiting
        responses = self.responses
        futures = self.futures
        frame = self.root
        send_msg = self.send_msg

        with self.log_as('futures_task') as ctx:
            self._check_loop(ctx)

            try:

                while len(awaiting) or len(futures):

                    # yield, so that we are only call recv_msg as last resort
                    await sleep(0)

                    self.__log_futures_task_status(ctx) if ctx else None

                    # might have finished during that time
                    if not len(awaiting) and not len(futures):
                        break

                    msg: RPCMsg = self.__recv_msg(ctx)

                    ctx.log('got msg:', msg) if ctx else None

                    if isinstance(msg, FramedRequestMsg):
                        exit_fn = self.on_frame(Incoming, msg)
                        if msg.is_async:
                            frame.exec_incoming_request_task(msg, loop, send_msg, exit_fn)
                        else:
                            frame.exec_incoming_request(msg, send_msg, exit_fn)

                    elif isinstance(msg, FramedResponseMsg):
                        mid = msg.mid
                        if future := awaiting.get(mid):
                            if not future.done():
                                ctx.log(f'future #{f_id(mid)} fulfilled:', future)
                                future.set_result(msg)
                                del awaiting[mid]
                            else:
                                ctx.warn(f'future #{f_id(mid)} already set', future, future._state)
                        else:
                            responses[mid] = msg

                    elif isinstance(msg, FutureEventMsg):
                        self.handle_future_event_msg(msg)

                    elif isinstance(msg, ChannelMsg):
                        channel = self.channel
                        if channel is not None:
                            channel.append(msg)

                    else:
                        self.raise_protocol_error(ctx, f'unexpected message: {msg}')

                ctx.log('ending futures task')

            finally:
                self.task = None

    ############################################################################

    def close(self):
        """Make a best-effort attempt to interrupt currently running msg_loop."""
        self.send_bytes = cast(SyncSendBytes, E.raise_shutdown)
        self.recv_bytes = cast(SyncRecvBytes, E.raise_shutdown)
        self.channel = None
        if futures := self.futures:
            for key, future in futures.items():
                self.unregister_future(future)
        super().close()

    ############################################################################

    def channel_close(self):
        self.send_msg(CHANNEL_CLOSED_MSG)

    def channel_send_value(self, value: Any, last: bool, /):
        channel_msg = ChannelMsg.encode(self.root, Result.good(value), last)
        self.send_msg(channel_msg)

    def channel_send_exception(self, exception: BaseException, last: bool, /):
        channel_msg = ChannelMsg.encode(self.root, Result.bad(exception), last)
        self.send_msg(channel_msg)

    def channel_recv(self) -> Result | Closed | Pending:
        queue = self.channel
        if queue is None:
            return CLOSED
        if not len(queue):
            return PENDING
        return self.__pop_queue()

    ############################################################################

    def remote_print(self, *args, sep: str = ' ', end: str = '\n', file=None, flush: bool = True):
        text = P._remote_print_str(args, sep=sep)
        self.send_msg(RemotePrintMsg(text))


NO_MID = 1 << 32



import asyncio
import dataclasses as DC

from agentica_internal.warpc.worlds.all import DebugWorld


async def run():
    async with DebugWorld.connected_pair(logging=False) as B:
        iv = DC.InitVar[int]
        iv_c = DC.InitVar[int]
        iv_b = await B(iv)
        print(iv, iv_b)
        print(iv == iv_b)


asyncio.run(run(), debug=True)



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.warpc.virtual_class_tests import verify_user_class, verify_generic_class
from agentica_internal.testing.warpc.virtual_dataclass_tests import verify_data_class, verify_empty_data_class


@pytest.mark.asyncio
async def test_warp_generic_class():
    await verify_generic_class()


@pytest.mark.asyncio
async def test_warp_user_class():
    await verify_user_class()


@pytest.mark.asyncio
async def test_warp_data_class():
    await verify_data_class()


@pytest.mark.asyncio
async def test_warp_empty_data_class():
    await verify_empty_data_class()



"""
This packages insulates us from actually having to *load* OpenTelemetry,
and also allows any of the Spans, Tracers, etc., to be None with no ill
effect.

* OSpan is an *optional* `opentelemetry.span.Span`
* OTracer is an *optional* `opentelemetry.tracer.Tracer`
* OAttributes is a dict of values allowed by opentelemetry
* OContextManager *could* be the context manager returned by
  `Tracer.start_as_current_span` whose `__enter__` returns a `span`,
  but it could also be a `contextlib.nullcontext` whose `__enter__` returns None.

`OContextManager` is returned by `tracer_span`, and you use it as:
    ```
    with tracer_span(tracer, 'name', ...) as span:
        span_set_attributes(span, ...)
        ...
    ```

When given a `tracer` is None, `span` will be None, otherwise it is
equivalent to `tracer.start_as_current_span(...)`.

The functions `span_set_attributes` etc. are the None-safe versions of
the ordinary methods on spans.
"""

from contextlib import AbstractContextManager, nullcontext
from typing import TYPE_CHECKING, Literal

__all__ = [
    'OSpan',
    'OTracer',
    'OContext',
    'OContextManager',
    'OAttributeValue',
    'OAttributes',
    'get_tracer',
    'span_set_context',
    'span_set_attributes',
    'span_set_attribute',
    'span_set_attributes',
    'span_set_error',
    'span_set_ok',
    'span_start',
    'span_finish',
    'span_end',
    'span_context_manager',
    'span_inject',
]

###############################################################################

if TYPE_CHECKING:
    from opentelemetry import trace
    from opentelemetry.util import types

###############################################################################

type OTracer = trace.Tracer | None
type OSpan = trace.Span | None
type OContext = trace.Context | None
type OAttributeValue = types.AttributeValue
type OAttributes = types.Attributes
type OContextManager = AbstractContextManager[OSpan]
type SpanKind = Literal['INTERNAL', 'SERVER', 'CLIENT', 'PRODUCER', 'CONSUMER']

###############################################################################


def get_tracer(name: str, enabled: bool) -> OTracer:
    if enabled and name:
        from opentelemetry.trace import get_tracer

        return get_tracer(name)
    return None


###############################################################################


def span_context_manager(tracer: OTracer, name: str, *args, **kwargs) -> OContextManager:
    if tracer is not None:
        return tracer.start_as_current_span(name, *args, **kwargs)
    return nullcontext()


def span_start(
    tracer: OTracer,
    name: str,
    context_span: OSpan = None,
    context: OContext = None,
    attributes: OAttributes = None,
    kind: SpanKind = 'INTERNAL',
) -> OSpan:
    if tracer is not None:
        from opentelemetry.trace import SpanKind, set_span_in_context

        span_kind = SpanKind[kind]
        if context_span is not None:
            context = set_span_in_context(context_span)
        return tracer.start_span(name, kind=span_kind, context=context, attributes=attributes)
    return None


def span_inject(span: OSpan, dct: dict) -> None:
    # Inject trace context into WebSocket headers for distributed tracing
    # This must happen AFTER setting the connection_span as the active span
    # so the correct trace context is propagated to the server
    if span is not None:
        from opentelemetry.propagate import inject
        from opentelemetry.trace import set_span_in_context

        context = set_span_in_context(span)
        inject(dct, context=context)


def span_finish(span: OSpan, error: BaseException | str | None = None) -> None:
    if span is not None:
        if error is not None:
            span_set_error(span, error)
        else:
            span_set_ok(span)
        span.end()


def span_end(span: OSpan) -> None:
    if span is not None:
        span.end()


################################################################################


def span_set_context(span: OSpan) -> OContext:
    if span is not None:
        from opentelemetry.trace import set_span_in_context

        return set_span_in_context(span)
    return None


def span_set_attribute(span: OSpan, name: str, value: OAttributeValue) -> None:
    if span is not None:
        span.set_attribute(name, value)


def span_set_attributes(span: OSpan, attributes: OAttributes) -> None:
    if span is not None and attributes:
        span.set_attributes(attributes)


def span_set_error(span: OSpan, error: BaseException | str | None = None) -> None:
    if span is not None:
        from opentelemetry.trace import Status, StatusCode

        if isinstance(error, BaseException):
            span.record_exception(error)
            error = str(error)
        elif type(error) is not str:
            error = None
        span.set_status(Status(StatusCode.ERROR, error))


def span_set_ok(span: OSpan) -> None:
    if span is not None:
        from opentelemetry.trace import Status, StatusCode

        span.set_status(Status(StatusCode.OK))


def span_add_link(span: OSpan, link: OSpan) -> None:
    if span is not None and link is not None:
        span.add_link(link.get_span_context())



# fmt: off

import asyncio

from .__ import *
from .base import *

__all__ = [
    'FutureData',
    'unregister_virtual_future'
]
from ...core.futures import HookableFuture


################################################################################

type FutureHandler = Callable[[FutureT], Any]

class FutureData(ResourceData):
    __slots__ = 'future', 'result'

    future: FutureT  # the existing real / novel virtual future created by `id_to_future`
    result: Result   # value/exception if done, PENDING_RESULT / CANCELED_RESULT if not

    # NOTE: this is pretty trivial, the reason is that the special sauce is
    # applied when a FutureData is *encoded* to a FutureDataMsg, and conversely
    # a FutureDataMsg is *decoded* to a FutureData. See FutureDataMsg in
    # `warpc.msg.resource_data` for that code.
    #
    # * during *encoding*, the future is registered with the owning world via
    # `frame.future_to_id`, which cooks up a future_id for it and attaches a
    # done_callback which will send events against this future_id.
    #
    # * during *decoding*, that future_id is passed to `frame.id_to_future` to create
    # matched local future which will respond to incoming `FutureEventMsgs` against
    # this future_id. This will be a HookableFuture to ensure that awaits and cancels
    # work correctly.
    #
    # see encode_fields and decode_fields in FutureDataMsg for where this happens

    @classmethod
    def describe_resource(cls, future: asyncio.Future) -> 'FutureData':
        data = FutureData()
        data.future = future
        data.result = Result.from_future(future)
        return data

    def create_resource(self, handle: ResourceHandle) -> FutureT:
        # this code ensures that the future behaves like a virtual resource,
        # triggering RPC when it is cancelled *locally*, the other code path, in
        # which it is set or cancelled via events received from the remote world,
        # happens because FutureDataMsg.decode_fields called `future_from_id` to
        # fill `data.future`
        future = self.future
        result = self.result
        handle.kind = Kind.Future
        setattr(future, VHDL, handle)
        result.into_future(future)  # no effect if result is pending
        register_virtual_future(future)
        return future

################################################################################

# this is only called by create_resource above!
def register_virtual_future(future: FutureT) -> None:

    if isinstance(future, HookableFuture):
        future.___set_hooks___(
            cancel_hook=send_future_cancel,
            set_result_hook=send_future_result,
            set_exception_hook=send_future_exception,
        )

# this should be called to DISABLE the triggering of RPC when a virtual future
# is canceled. it is called by `send_future_request`, but ALSO by a Frame when it
# receives a future event and sets the virtual future's outcome to match what
# happened with the remote future, to avoid the close event from triggering RPC.
def unregister_virtual_future(future: FutureT) -> None:

    if isinstance(future, HookableFuture):
        future.___del_hooks___()

# these functions allow *local* calls to .set_result, etc. for a matched future
# to trigger messages that will cancel/set the original future via a *framed*
# ResourceRequest using the same 'handle @' mechanism used for other virtual
# resources.
def send_future_cancel(future: asyncio.Future, _: None) -> None:

    unregister_virtual_future(future)

    if handle := get_handle(future):
        event = CancelFuture(future)
        handle.hdlr(handle, event)


def send_future_result(future: asyncio.Future, result: Any) -> None:

    unregister_virtual_future(future)

    if handle := get_handle(future):
        event = CompleteFuture(future, Result.good(result))
        handle.hdlr(handle, event)


def send_future_exception(future: asyncio.Future, exception: BaseException) -> None:

    unregister_virtual_future(future)

    if handle := get_handle(future):
        event = CompleteFuture(future, Result.bad(exception))
        handle.hdlr(handle, event)



# fmt: off

from typing import Callable

from agentica_internal.cpython.frame import FrameType, exception_frame

__all__ = [
    'unary_user_fn_frame',
    'variadic_user_fn_frame',
    'FRAMES',
]


def get_exception_frame(fn: Callable, *args, **kwargs) -> FrameType:
    try:
        fn(*args, **kwargs)
    except Exception as e:
        frame = exception_frame(e)
        assert frame
        return frame
    assert False


def unary_user_fn(a: int) -> bool:
    raise Exception()
    return True


def variadic_user_fn(*args: int, **kwargs: str) -> bool:
    raise Exception()
    return True


unary_user_fn_frame = get_exception_frame(unary_user_fn, 5)
variadic_user_fn_frame = get_exception_frame(variadic_user_fn, 1, 2, 3, a='x', b='y')


FRAMES = [unary_user_fn_frame, variadic_user_fn_frame]



################################################################################

import sys
from types import MappingProxyType
from typing import Any, Iterable, ClassVar, Iterator
from collections.abc import Iterable, Iterator

################################################################################

__classes__ = ['dict_keys', 'dict_values', 'dict_items']

__functions__ = []

__aliases__ = []

__annos__ = []

__constants__ = []

__all__ = ['dict_keys', 'dict_values', 'dict_items']

################################################################################


class dict_keys:
    __hash__: ClassVar[None]

    def __eq__(self, value: object, /) -> bool: ...
    def __reversed__(self) -> Iterator: ...

    if sys.version_info >= (3, 13):

        def isdisjoint(self, other: Iterable, /) -> bool: ...

    if sys.version_info >= (3, 10):

        @property
        def mapping(self) -> MappingProxyType: ...


################################################################################


class dict_values:
    def __reversed__(self) -> Iterator: ...

    if sys.version_info >= (3, 10):

        @property
        def mapping(self) -> MappingProxyType: ...


################################################################################


class dict_items:
    __hash__: ClassVar[None]

    def __eq__(self, value: object, /) -> bool: ...
    def __reversed__(self) -> Iterator[tuple[Any, Any]]: ...

    if sys.version_info >= (3, 13):

        def isdisjoint(self, other: Iterable[tuple[Any, Any]], /) -> bool: ...

    if sys.version_info >= (3, 10):

        @property
        def mapping(self) -> MappingProxyType: ...



import asyncio

from agentica_internal.warpc.forbidden import *
from agentica_internal.warpc.worlds.all import DebugWorld


class Foo:
    w: DebugWorld
    i: int

    def getw(self, s: str) -> DebugWorld:
        return self.w


class Bar:
    i: int


async def run():
    w = DebugWorld('z')

    foo = Foo()
    foo.w = w
    foo.i = 4

    bar = Bar()
    bar.i = 5

    async with DebugWorld.connected_pair(logging=False) as B:
        bar_b = await B(bar)

        print("\n\nsending DebugWorld instance")
        w_b = await B(w)
        print(w, w_b)

        print("\n\nsending DebugWorld class")
        DebugWorld_b = await B(DebugWorld)
        print(DebugWorld, DebugWorld_b)

        print("\n\nsending Foo")
        Foo_b = await B(Foo)
        print(Foo.__annotations__, Foo_b.__annotations__)


asyncio.run(run(), debug=True)



import asyncio
import typing

from agentica_internal.testing.examples.overloads import foo
from agentica_internal.warpc.worlds.all import DebugWorld


async def run():
    async with DebugWorld.connected_pair(logging=False) as B:
        foo_b = await B(foo)

        foo_overloads = typing.get_overloads(foo)
        foo_b_overloads = typing.get_overloads(foo_b)

        assert len(foo_overloads) == len(foo_b_overloads)

        print(f'overloads for foo ({foo=!r})')
        for o in foo_overloads:
            print(o.__name__, o, type(o), o.__doc__)

        print(f'overloads for foo_b ({foo_b=!r})')
        for o in foo_b_overloads:
            print(o.__name__, o, type(o), o.__doc__)

        for o, o_b in zip(foo_overloads, foo_b_overloads):
            assert o != o_b
            assert o.__name__ == o_b.__name__
            assert o.__doc__ == o_b.__doc__
            assert o.__annotations__ == o_b.__annotations__


asyncio.run(run(), debug=True)



# fmt: off

from types import FunctionType
from typing import overload

__all__ = [
    "fn_nullary",
    "fn_unary",
    "fn_binary",
    "fn_unary_1_def",
    "fn_binary_1_def",
    "fn_binary_2_def",
    "fn_with_annos",
    "fn_with_star",
    "fn_with_dstar",
    "fn_with_star_dstar",
    "fn_blank",
    "fn_pos_only",
    "fn_kw_only",
    "fn_pos_and_kw_only",
    "fn_overloaded",
    "fn_complex",
    "fn_async",
    "fn_async_gen",

    "FUNCTIONS",
]

type MyType = list[str] | None


def fn_nullary(): ...
def fn_unary(a1): ...
def fn_binary(a1, a2): ...


def fn_unary_1_def(a1: int = 0): ...
def fn_binary_1_def(a1: int, a2: bool = True) -> None: ...
def fn_binary_2_def(a1: int = 0, a2: bool = True) -> None: ...


def fn_with_annos(a1: int, a2: float) -> MyType: ...


def fn_with_star(a1: int, *args: str) -> None: ...
def fn_with_dstar(a1: int, **kwargs: str) -> None: ...
def fn_with_star_dstar(a1: int, *args: str, **kwargs: str) -> None: ...
def fn_blank(*args, **kwargs): ...


def fn_pos_only(a1_po: str, /, a2: bool, a3: str) -> None: ...
def fn_kw_only(a1: str, *, a2_kw: bool, a3_kw: str) -> None: ...
def fn_pos_and_kw_only(a1_po: str, /, a2: bool, *, a3_kw: str) -> None: ...


@overload
def fn_overloaded() -> int: ...
@overload
def fn_overloaded(a1: int) -> bool: ...
def fn_overloaded(a1: int = 0) -> int | bool: ...


def fn_complex(
    a1: int, a2: bool, /, a3: str, *args: str, a4: bool = True, **kwargs: str
) -> None: ...


async def fn_async(a1: int):
    return


async def fn_async_gen(a1: int):
    yield None


FUNCTIONS: list[FunctionType] = [
    len,
    fn_nullary,
    fn_unary,
    fn_binary,
    fn_unary_1_def,
    fn_binary_1_def,
    fn_binary_2_def,
    fn_with_annos,
    fn_with_star,
    fn_with_dstar,
    fn_with_star_dstar,
    fn_blank,
    fn_pos_only,
    fn_kw_only,
    fn_pos_and_kw_only,
    fn_overloaded,
    fn_complex,
    fn_async,
    fn_async_gen,
]



# fmt: off

from .interface import (
    AsyncRecvBytes,
    AsyncSendBytes,
    SyncRecvBytes,
    SyncSendBytes,
)

from .base_world import World
from .sdk_world import SDKWorld
from .agent_world import AgentWorld
from .debug_world import DebugWorld

__all__ = [
    'World',
    'AgentWorld',
    'SDKWorld',
    'SyncSendBytes',
    'SyncRecvBytes',
    'AsyncSendBytes',
    'AsyncRecvBytes',
    'DebugWorld',
]



from types import ModuleType

from agentica_internal.core.mixin import finalize


def transform(mod: ModuleType) -> None:
    for name, value in mod.__dict__.items():
        if isinstance(value, int):
            mod.__dict__[name] = value * 10


finalize(transform)


x = 2
y = 3



# fmt: off

from ..classes.dclasses import DCLASS_INSTANCES
from ..classes.diamonds import DIAMOND_INSTANCES
from ..classes.enums import ENUM_INSTANCES
from ..classes.named_tuples import NAMED_TUPLE_INSTANCES
from ..classes.typed_dicts import TYPED_DICT_INSTANCES
from ..classes.with_annos import WITH_ANNO_INSTANCES
from ..classes.with_methods import WITH_METHOD_INSTANCES
from ..classes.with_props import WITH_PROP_INSTANCES

__all__ = [
    'USER_CLASS_INSTANCES',
]

USER_CLASS_INSTANCES = [
    *DCLASS_INSTANCES,
    *DIAMOND_INSTANCES,
    *ENUM_INSTANCES,
    *NAMED_TUPLE_INSTANCES,
    *TYPED_DICT_INSTANCES,
    *WITH_ANNO_INSTANCES,
    *WITH_METHOD_INSTANCES,
    *WITH_PROP_INSTANCES,
]



from types import FunctionType
from typing import Any, NamedTuple, TypeGuard

__all__ = [
    'is_function',
    'func_location',
    'func_flags',
    'func_arity',
    'func_defaults',
    'func_arg_names',
    'func_arg_strs',
    'func_sig_info',
    'FuncFlags',
    'FuncSigInfo',
    'FuncArgInfo',
]


def func_args_with_default(func: FunctionType) -> dict[str, Any]:
    """Get dictionary of argument names to their default value if specified."""
    defaults: dict[str, Any] = dict()
    if func.__defaults__:
        varnames = list(func.__code__.co_varnames)[: func.__code__.co_argcount][
            -len(func.__defaults__) :
        ]
        defaults |= dict(zip(varnames, func.__defaults__))
    if func.__kwdefaults__:
        defaults |= func.__kwdefaults__
    return defaults


def is_function(obj: object) -> TypeGuard[FunctionType]:
    """Returns True for a `FunctionType` object: functions explicitly declared with `def foo(...)`."""

    return isinstance(obj, FunctionType)


def func_location(func: FunctionType) -> str:
    assert isinstance(func, FunctionType)

    code = func.__code__
    return f'{code.co_filename}#{code.co_firstlineno}'


# TODO: rename to `func_simple_arg_count`
def func_arity(func: FunctionType) -> int:
    """
    Returns the number of *simple* arguments of `FunctionType` object.
    Use `func_arg_info` to get arity accounting for `*args` and `**kwargs`.

    Examples:
    `def fn(a, b)`             -> 2
    `def fn(a, b, *args)`      -> 2
    `def fn(a, b, **kwargs)`   -> 2
    `def fn(a, b, /, foo=bar)` -> 2
    """

    code = func.__code__
    return code.co_argcount


def func_defaults(func: FunctionType) -> dict[str, Any]:
    """Returns a dictionary mapping argument names to default values for a `FunctionType` object."""

    defaults = func.__defaults__
    if defaults is None:
        return {}
    arg_names = func_arg_names(func)
    num = len(defaults)
    return dict(zip(arg_names[-num:], defaults))


# TODO: this doesn't include *args and **kwargs, but should it?
def func_arg_names(func: FunctionType) -> tuple[str, ...]:
    """Given a FunctionType object, return the tuple of *simple* arg names,

    Examples:
    `def fn(a, b)`             -> ('a', 'b')
    `def fn(a, b, *args)`      -> ('a', 'b')
    `def fn(a, b, **kwargs)`   -> ('a', 'b')
    `def fn(a, b, /, foo=bar)` -> ('a', 'b')
    """

    code = func.__code__
    arg_count = code.co_argcount
    var_names = code.co_varnames
    return var_names[:arg_count]


def func_arg_strs(func: FunctionType) -> list[str]:
    """
    Given a `FunctionType` object, returns tuple of argument strings that would appear in
    its `def fn(...)` form, including '/', '*', '*args', and '**kwargs'.
    """
    from intro.cpython.code import code_arg_info

    assert type(func) is FunctionType, f'{func} is not FunctionType'
    code = func.__code__
    vars, num_pos, num_reg, num_key, pos_star, key_star = code_arg_info(code)
    args = list(vars)

    strs = []
    add = strs.append
    ext = strs.extend
    pop = args.pop

    def pull(n: int):
        ext(pop(0) for _ in range(n))

    if num_pos:
        pull(num_pos)
        add('/')
    if num_reg:
        pull(num_reg)
    if num_key:
        add('*')
        pull(num_key)
    if pos_star:
        add('*' + pop())
    if key_star:
        add('**' + pop())

    return strs


class FuncFlags(NamedTuple):
    """Represents flags about a function:

    | 0 | `has_defaults` | True if any positional or key args have defaults
    | 1 | `has_pos_only` | True if any arguments are positional only
    | 2 | `has_key_only` | True if any arguments are keyword only
    | 3 | `has_pos_star` | True if there is a `*args` argument
    | 4 | `has_key_star` | True if there is a `**kwargs` argument
    """

    has_defaults: bool
    has_pos_only: bool
    has_key_only: bool
    has_pos_star: bool
    has_key_star: bool

    def is_variadic(self) -> bool:
        return self.has_defaults or self.has_key_star or self.has_pos_star

    def is_simple(self) -> bool:
        return not any(a for a in self)


def func_flags(func: FunctionType) -> FuncFlags:
    """Given a `FunctionType` object, returns a set of flags about it."""

    assert type(func) is FunctionType, f'{func} is not FunctionType'

    has_defaults = bool(func.__defaults__) or bool(func.__kwdefaults__)
    from .code import code_arg_info

    info = code_arg_info(func.__code__)

    return FuncFlags(has_defaults, info.num_pos > 0, info.num_key > 0, info.pos_star, info.key_star)


class NoDefault:
    def __repr__(self):
        return 'NO_DEFAULT'

    __str__ = __repr__


NO_DEFAULT = NoDefault()


class FuncArgInfo(NamedTuple):
    """Represents information about a function argument:

    | 0 | `name`     | `str`    | argument name
    | 1 | `required` | `bool`   | False if there is a default value
    | 2 | `default`  | `object` | default value, or `FuncArgInfo.NO_DEFAULT` if no default
    | 3 | `pos_only` | `bool`   | True if the argument is positional-only
    | 4 | `key_only` | `bool`   | True if the argument is keyword-only
    """

    name: str
    optional: bool
    default: object
    pos_only: bool
    key_only: bool
    NO_DEFAULT = NO_DEFAULT

    def __repr__(self) -> str:
        flags = ''
        if self.optional:
            flags += 'opt '
        if self.pos_only:
            flags += 'pos_only '
        if self.key_only:
            flags += 'key_only '
        return f"<{flags}arg {self.name!r}>"


class FuncSigInfo(NamedTuple):
    """Represents information about a function signature:

    | 0 | `fn_name`     | (qualified) name of the function itself
    | 1 | `arg_names`   | tuple of argument names (excluding star args)
    | 2 | `pos_star`    | name of '*' argument, or None
    | 3 | `key_star`    | name of '**' argument, or None
    | 4 | `min_args`    | min arg count
    | 5 | `max_args`    | max arg count, or -1 if no limit
    | 6 | `arg_info`    | list of FuncArgInfo tuples describing non-star args
    | 7 | `is_async`    | True if this is an `async def` function
    """

    fn_name: str
    arg_names: tuple[str, ...]
    pos_star: str | None
    key_star: str | None
    min_args: int
    max_args: int
    arg_info: tuple[FuncArgInfo, ...]
    is_async: bool

    def __repr__(self) -> str:
        args = list(map(repr, self.arg_info))
        if self.pos_star:
            args.append(f'star={self.pos_star!r}')
        if self.key_star:
            args.append(f'dstar={self.key_star!r}')
        f_args = ', '.join(args)
        return f"<sig ({f_args}) for {self.fn_name!r}>"

    def opt_args(self) -> tuple[str, ...]:
        return tuple(arg.name for arg in self.arg_info if arg.default is not NO_DEFAULT)

    def all_arg_names(self) -> list[str]:
        arg_names = list(self.arg_names)
        if pos_star := self.pos_star:
            arg_names.append(pos_star)
        if key_star := self.key_star:
            arg_names.append(key_star)
        return arg_names


def func_sig_info(func: FunctionType) -> FuncSigInfo:
    """Given a FunctionType object, returns a FuncSigInfo object."""

    from agentica_internal.cpython.code import FLAGS, code_arg_info

    assert type(func) is FunctionType, f'{func} is not FunctionType object'
    fn_code = func.__code__
    pos_defs = func.__defaults__ or ()
    key_defs = func.__kwdefaults__ or {}
    fn_name = func.__qualname__

    var_names, num_pos, num_reg, num_key, pos_star, key_star = code_arg_info(fn_code)
    is_async = bool(fn_code.co_flags & FLAGS.IS_ASYNC)
    first_key = num_pos + num_reg
    first_star = first_key + num_key
    first_def = first_key - len(pos_defs)
    arg_names = var_names[:first_star]

    arg_info: list[FuncArgInfo] = []
    add_arg = arg_info.append
    min_args = max_args = 0
    for i in range(first_star):
        name = var_names[i]
        key_only = i >= first_key
        pos_only = i < num_pos
        if key_only:
            optional = name in key_defs
            default = key_defs[name] if optional else NO_DEFAULT
        else:
            optional = i >= first_def
            default = pos_defs[i - first_def] if optional else NO_DEFAULT
        min_args += 0 if optional else 1
        max_args += 1
        add_arg(FuncArgInfo(name, optional, default, pos_only, key_only))

    def get_star_name():
        nonlocal first_star, max_args
        max_args = -1
        first_star += 1  # type: ignore
        return var_names[first_star - 1]

    pos_star_name = get_star_name() if pos_star else None
    key_star_name = get_star_name() if key_star else None

    return FuncSigInfo(
        fn_name,
        arg_names,
        pos_star_name,
        key_star_name,
        min_args,
        max_args,
        tuple(arg_info),
        is_async,
    )
